import logging
from datetime import datetime as pydt

import numpy as np
import pandas as pd
import xarray as xr

from wam2layers import __version__
from wam2layers.config import Config
from wam2layers.preprocessing.shared import add_bounds
from wam2layers.reference.variables import DATA_ATTRIBUTES

logger = logging.getLogger(__name__)


def input_path(date, config):
    input_dir = config.preprocessed_data_folder
    return f"{input_dir}/{date.strftime('%Y-%m-%d')}_fluxes_storages.nc"


def output_path(date, config):
    output_dir = config.output_folder
    mode = "backtrack" if config.tracking_direction == "backward" else "forwardtrack"
    return f"{output_dir}/{mode}_{date.strftime('%Y-%m-%dT%H-%M')}.nc"


def read_data_at_date(d, config):
    """Load input data for given date."""
    file = input_path(d, config)
    ds = xr.open_dataset(file, cache=False)
    if config.tracking_domain is not None:
        return select_subdomain(ds, config.tracking_domain)
    return ds


def read_data_at_time(t, config):
    """Get a single time slice from input data at time t."""
    ds = read_data_at_date(t, config)
    return ds.sel(time=t, drop=True)


def select_subdomain(ds, bbox):
    """Limit the data to a subdomain."""
    if bbox.west < bbox.east:
        return ds.sel(
            longitude=slice(bbox.west, bbox.east),  # -180 to 180
            latitude=slice(bbox.north, bbox.south),  # 90 - -90 (decreasing)
            drop=True,
        )
    else:
        # Need to roll coordinates to maintain a continuous longitude
        # Roll the first lon that's NOT in the bbox to the front of the array
        # That way, all values that ARE in the bbox, must be contiguous.
        in_bbox = (ds.longitude < bbox.east) | (ds.longitude > bbox.west)
        n_roll = in_bbox.argmin().item()  # first false value
        ds_rolled = ds.roll(longitude=-n_roll, roll_coords=True)
        in_bbox = (ds_rolled.longitude < bbox.east) | (ds_rolled.longitude > bbox.west)
        return ds_rolled.sel(longitude=in_bbox, latitude=slice(bbox.north, bbox.south))


def load_data(t, config, subset="fluxes"):
    """Load variable at t, interpolate if needed."""
    variables = {
        "fluxes": ["fx_upper", "fx_lower", "fy_upper", "fy_lower", "evap", "precip"],
        "states": ["s_upper", "s_lower"],
    }

    t1 = t.ceil(config.input_frequency)
    da1 = read_data_at_time(t1, config)[variables[subset]]
    if t == t1:
        # this saves a lot of work if times are aligned with input
        return da1

    t0 = t.floor(config.input_frequency)
    da0 = read_data_at_time(t0, config)[variables[subset]]
    if t == t0:
        return da0

    return da0 + (t - t0) / (t1 - t0) * (da1 - da0)


def load_tagging_region(config, t=None):
    tagging_region = xr.open_dataarray(config.tagging_region)

    if config.tracking_domain is not None:
        tagging_region = select_subdomain(tagging_region, config.tracking_domain)

    if t is not None:
        return tagging_region.sel(time=t, method="nearest")
    return tagging_region


def add_time_bounds(ds: xr.Dataset, t: pd.Timestamp, config: Config):
    """Compute the time bounds and add them to the output dataset."""
    dt = pd.Timedelta(config.output_frequency)
    if config.tracking_direction == "forward":
        ds["time_bnds"] = (("time", "bnds"), np.array(((t - dt, t),)))
    else:
        ds["time_bnds"] = (("time", "bnds"), np.array(((t + dt, t),)))


def get_main_attrs(attrs: dict, config: Config):
    """Create the main dataset attributes, including appending to the history."""
    new_history = (
        f"created on {pydt.utcnow():%Y-%m-%dT%H:%M:%SZ} "
        f"using wam2layers version {__version__}; "
    )
    return {
        "title": f"WAM2Layers {config.tracking_direction}tracking output file",
        "history": new_history + attrs["history"]
        if "history" in attrs
        else new_history,
        "source": (
            "Moisture tracking applied to ERA5 dataset: ECMWF Reanalysis v5 (ERA5), "
            "www.ecmwf.int/en/forecasts/dataset/ecmwf-reanalysis-v5"
        ),
        "references": "doi.org/10.5281/zenodo.7010594, doi.org/10.5194/esd-5-471-2014",
        "Conventions": "CF-1.6",
        "WAM2Layers_config": config.model_dump_json(),
    }


def write_output(
    output: xr.Dataset,
    t: pd.Timestamp,
    config: Config,
) -> None:
    # TODO: add back (and cleanup) coordinates and units
    path = output_path(t, config)
    logger.info(f"{t} - Writing output to file {path}")

    output = output.assign_coords({"time": t})
    output = output.expand_dims(dim="time")

    for var in output.data_vars:
        output[var].attrs.update(DATA_ATTRIBUTES[str(var)])

    add_bounds(output)
    add_time_bounds(output, t, config)
    output.attrs = get_main_attrs(output.attrs, config)

    comp = dict(zlib=True, complevel=8)
    encoding = {var: comp for var in output.data_vars}
    time_encoding = {"units": "seconds since 1900-01-01"}
    encoding["time"] = time_encoding
    encoding["time_bnds"] = time_encoding

    for var in output.data_vars:
        if var != "time_bnds":
            output[var] = output[var].astype("float32")
    output.to_netcdf(path, encoding=encoding)



from functools import partial

import numpy as np

from wam2layers.config import Config
from wam2layers.utils.profiling import ProgressTracker


def pad_boundaries(*args, periodic=False):
    """Add boundary padding to all input arrays.

    Arguments:
        *args: Input arrays to which boundary padding will be added.
        periodic: If True, apply periodic boundary conditions when padding in
            the x-direction. If False, pad with zeros in the x and y directions.

    Returns:
        List of input arrays with boundary padding added.
    """
    zero_xy = partial(np.pad, pad_width=1, mode="constant", constant_values=0)
    periodic_x = partial(np.pad, pad_width=((0, 0), (1, 1)), mode="wrap")
    zero_y = partial(
        np.pad, pad_width=((1, 1), (0, 0)), mode="constant", constant_values=0
    )

    if periodic:
        return [periodic_x(zero_y(arg)) for arg in args]
    return [zero_xy(arg) for arg in args]


def horizontal_advection(s, u, v, periodic_x=False) -> np.ndarray:
    """Calculate advection on a staggered grid using a donor cell scheme.

    Boundaries are padded with 0 or periodic boundaries to maintain the original
    grid size.

    The advection equation reads `grad(u*s)` where u is the velocity vector and
    s is any scalar. Using i for grid cell indices in the x direction and j for
    the y-direction (increasing northwards), the donor cell scheme reads:

    s(i-1/2) = s(i-1) if u(i-1/2) > 0
               s(i+1) otherwise
    s(j-1/2) = s(j-1) if v(j-1/2) > 0
               s(j+1) otherwise

    d(us)/dx = u(i-1/2)*s(i-1/2) - u(i+1/2)*s(i+1/2)
    d(vs)/dy = v(j-1/2)*s(j-1/2) - v(j+1/2)*s(j+1/2)

    adv(s) = d(us)/dx + d(vs)/dy

    Arguments:
        s: array of shape [M, N]. It is assumed that the grid dimensions are
            [latitude, longitude] and latitude is in decreasing order.
        u: array of shape = [M-2, N-1]
        v: array of shape = [M-1, N-2]

    Returns:
        array of shape [M, N]

    Examples:

        Create a simple array:
        >>> s = np.zeros((5, 5))
        >>> s[2, 2] = 1
        >>> u = np.ones((3, 4)) * .5
        >>> v = np.ones((4, 3)) * .5
        >>> s
        array([[0., 0., 0., 0., 0.],
               [0., 0., 0., 0., 0.],
               [0., 0., 1., 0., 0.],
               [0., 0., 0., 0., 0.],
               [0., 0., 0., 0., 0.]])

        Calculate the advection for a single time step (forward)
        >>> horizontal_advection(s, u, v)
        array([[ 0. ,  0. ,  0. ,  0. ,  0. ],
               [ 0. ,  0. ,  0.5,  0. ,  0. ],
               [ 0. ,  0. , -1. ,  0.5,  0. ],
               [ 0. ,  0. ,  0. ,  0. ,  0. ],
               [ 0. ,  0. ,  0. ,  0. ,  0. ]])

        Backtracking can be done by reversing the velocities:
        >>> horizontal_advection(s, -u, -v)
        array([[ 0. ,  0. ,  0. ,  0. ,  0. ],
               [ 0. ,  0. ,  0. ,  0. ,  0. ],
               [ 0. ,  0.5, -1. ,  0. ,  0. ],
               [ 0. ,  0. ,  0.5,  0. ,  0. ],
               [ 0. ,  0. ,  0. ,  0. ,  0. ]])

    """
    # Pad boundaries with ghost cells
    qp, up, vp = pad_boundaries(s, u, v, periodic=periodic_x)
    # shapes: [M+2, N+2], [M, N+1], [M+1, N]

    west = np.s_[:-1]
    east = np.s_[1:]
    south = np.s_[1:]
    north = np.s_[:-1]
    inner = np.s_[1:-1]

    # Donor cell upwind scheme (2 directions seperately)
    uq = np.where(up > 0, up * qp[inner, west], up * qp[inner, east])  # [M, N+1]

    vq = np.where(vp > 0, vp * qp[south, inner], vp * qp[north, inner])  # [M, N+2]

    adv_x = uq[:, west] - uq[:, east]  # [M, N]
    adv_y = vq[south, :] - vq[north, :]  # [M, N]

    return adv_x + adv_y  # [M, N]


def vertical_advection(fv, s_lower, s_upper):
    """Calculate 1d upwind advection of vertical flux.

    Upwind advection with fv positive downwards, so:

    fv * s = fv * s_upper if fv > 0
           = fv * s_lower otherwise
    """
    return np.where(fv >= 0, fv * s_upper, fv * s_lower)


def vertical_dispersion(fv, s_lower, s_upper, kvf):
    """Calculate additional vertical mixing due to convective dispersion.

    dispersion = kvf * |Fv| * dS/dz
    """
    return kvf * np.abs(fv) * (s_upper - s_lower)


def stabilize_fluxes(F, S, progress_tracker: ProgressTracker, config: Config, t):
    """Stabilize the outfluxes / influxes.

    CFL: Water cannot move further than one grid cell per timestep.

    Arguments:
        F: xr.Dataset holding the fluxes
        S: xr.Dataset holding the states
        progress_tracker: ProgressTracker object for writing log messages
        config: Config object with the configuration for this experiment
        t: current time (for writing output)
    """
    for level in ["upper", "lower"]:
        fx = F["fx_" + level] * config.timestep
        fy = F["fy_" + level] * config.timestep
        s = S["s_" + level]

        fx_abs = np.abs(fx)
        fy_abs = np.abs(fy)
        ft_abs = fx_abs + fy_abs

        # TODO: make 1/2 configurable?
        fx_limit = 1 / 2 * fx_abs / ft_abs * s.values
        fx_stable = np.minimum(fx_abs, fx_limit)

        fy_limit = 1 / 2 * fy_abs / ft_abs * s.values
        fy_stable = np.minimum(fy_abs, fy_limit)

        progress_tracker.track_stability_correction(fy_stable, fy_abs, config, t)

        # Get rid of any nan values
        fx_stable.fillna(0)
        fy_stable.fillna(0)

        # Re-instate the sign and convert back to flux instead of accumulation
        F["fx_" + level] = np.sign(fx) * fx_stable / config.timestep
        F["fy_" + level] = np.sign(fy) * fy_stable / config.timestep


def divergence(fx, fy):
    # Note: latitude decreasing, hence negative fy gradient is divergence
    return np.gradient(fx, axis=-1) - np.gradient(fy, axis=-2)


def calculate_fz(F, S0, S1, dt, kvf):
    """Calculate the vertical fluxes.

    The vertical flux is calculated as a closure term. Residuals are distributed
    proportionally to the amount of moisture in each layer.

    The flux is constrained such that it can never exceed

    Arguments:
        F: xarray dataset with fluxes evaluated at temporal midpoints between states
        S0: xarray dataset with states at current time t
        S1: xarray dataset with states at updated time t+1 (always forward looking)
        dt: timestep
        kvf: net to gross vertical flux multiplication parameter. With the a value of 0,
            the net and gross fluxes are equal.

    Returns:
        fz: vertical flux, positive downward

    Examples:

        Create dummy input data. In the absence of any horizontal fluxes and
        sources or sinks, this result can be verified manually.

        >>> import numpy as np
        >>> import xarray as xr
        >>> from wam2layers.preprocessing.shared import calculate_fz
        >>>
        >>> F = xr.Dataset({
        ...     'precip': xr.DataArray(np.zeros(((3, 3))), dims=['lat', 'lon']),
        ...     'evap': xr.DataArray(np.zeros(((3, 3))), dims=['lat', 'lon']),
        ...     'fx_upper': xr.DataArray(np.zeros(((3, 3))), dims=['lat', 'lon']),
        ...     'fy_upper': xr.DataArray(np.zeros(((3, 3))), dims=['lat', 'lon']),
        ...     'fx_lower': xr.DataArray(np.zeros(((3, 3))), dims=['lat', 'lon']),
        ...     'fy_lower': xr.DataArray(np.zeros(((3, 3))), dims=['lat', 'lon']),
        ... })
        >>> S0 = xr.Dataset({
        ...     's_upper': 10 * xr.DataArray(np.ones(((3, 3))), dims=['lat', 'lon']),
        ...     's_lower': 6 * xr.DataArray(np.ones(((3, 3))), dims=['lat', 'lon']),
        ... })
        >>> S1 = xr.Dataset({
        ...     's_upper': 8 * xr.DataArray(np.ones(((3, 3))), dims=['lat', 'lon']),
        ...     's_lower': 7 * xr.DataArray(np.ones(((3, 3))), dims=['lat', 'lon']),
        ... })
        >>> calculate_fz(F, S0, S1, dt=1)
        <xarray.DataArray (lat: 3, lon: 3)>
        array([[1.46666667, 1.46666667, 1.46666667],
               [1.46666667, 1.46666667, 1.46666667],
               [1.46666667, 1.46666667, 1.46666667]])
        Dimensions without coordinates: lat, lon

    """
    s_mean = (S1 + S0) / 2
    s_total = s_mean.s_upper + s_mean.s_lower
    s_rel = s_mean / s_total

    # Evaluate all terms in the moisture balance execpt the unknown Fz and err
    foo_upper = (S1 - S0).s_upper + dt * (
        divergence(F.fx_upper, F.fy_upper) + F.precip.values * s_rel.s_upper
    )
    foo_lower = (S1 - S0).s_lower + dt * (
        divergence(F.fx_lower, F.fy_lower) + F.precip.values * s_rel.s_lower - F.evap
    )

    # compute the resulting vertical moisture flux; the vertical velocity so
    # that the new err_lower/s_lower = err_upper/s_upper (positive downward)
    fz = foo_lower - S1.s_lower / (S1.s_lower + S1.s_upper) * (foo_upper + foo_lower)

    # TODO: verify that err_lower/s_lower = err_upper/s_upper is satisfied on debug log

    # stabilize the outfluxes / influxes; during the reduced timestep the
    # vertical flux can maximally empty/fill 1/x of the top or down storage
    stab = 1.0 / (kvf + 1.0)
    flux_limit = np.minimum(
        s_mean.s_upper, s_mean.s_lower
    )  # TODO why is this not 'just' the upstream bucket?
    fz_stable = np.minimum(np.abs(fz), stab * flux_limit)

    # Reinstate the sign and convert accumulation back to flux
    return np.sign(fz) * fz_stable / dt



import logging
from datetime import datetime

import pytest

from wam2layers.config import BoundingBox, Config

EXAMPLE_CONFIG = "example_config.yaml"
TEST_CONFIG = "tests/test_data/config_rhine.yaml"


def test_load_example_config():
    """Verify that the example config file can be loaded"""
    Config.from_yaml(EXAMPLE_CONFIG)


def test_load_test_config():
    """Verify that the test config file can be loaded"""
    Config.from_yaml(TEST_CONFIG)


@pytest.mark.parametrize("config_file", [EXAMPLE_CONFIG, TEST_CONFIG])
def test_config_export(tmp_path, config_file):
    """Verify that a config written to a file is identical to the original one."""
    cfg = Config.from_yaml(EXAMPLE_CONFIG)
    export_path = tmp_path / "config.yaml"
    cfg.to_file(export_path)
    cfg_exported = Config.from_yaml(export_path)
    assert cfg_exported == cfg


def test_check_date_order_invalid_tracking():
    """Raise error when tracking_end_date is earlier than tracking_start_date."""
    config = Config.from_yaml(TEST_CONFIG)
    with pytest.raises(
        ValueError, match="tracking_end_date should be later than tracking_start_date"
    ):
        config.tracking_end_date = datetime(2022, 7, 31)


def test_check_date_order_invalid_tagging():
    """Raise error when tagging_end_date is earlier than tagging_start_date."""
    config = Config.from_yaml(TEST_CONFIG)
    with pytest.raises(
        ValueError, match="tagging_end_date should be later than tagging_start_date"
    ):
        config.tagging_end_date = datetime(2022, 7, 31)


def test_check_date_order_invalid_preprocess():
    """Raise error when preprocess_end_date is earlier than preprocess_start_date."""
    config = Config.from_yaml(TEST_CONFIG)
    with pytest.raises(
        ValueError,
        match="preprocess_end_date should be later than preprocess_start_date",
    ):
        config.preprocess_end_date = datetime(2022, 7, 31)


class TestValidateRegion:
    def test_path_expanded(self):
        """Check path is expanded."""
        config = Config.from_yaml(TEST_CONFIG)
        assert config.tagging_region.is_absolute()
        config.tagging_region = "tests/test_data/region_rhine.nc"
        assert config.tagging_region.is_absolute()

    def test_invalid_path(self):
        """Raise error for invalid path."""
        config = Config.from_yaml(TEST_CONFIG)
        with pytest.raises(Exception, match="does not point to a file"):
            config.tagging_region = "weird/path.exe"

    def test_valid_bbox_ok(self):
        """Check that we can set region to a bbox"""
        config = Config.from_yaml(TEST_CONFIG)
        config.tagging_region = BoundingBox(0, 50, 10, 55)
        assert config.tagging_region == BoundingBox(0, 50, 10, 55)

    def test_invalid_south(self):
        """Raise error for invalid bbox."""
        config = Config.from_yaml(TEST_CONFIG)
        with pytest.raises(Exception, match="south should be smaller than north"):
            config.tagging_region = BoundingBox(0, 60, -10, 55)

    def test_wrap_meridian(self, caplog):
        """Warn for bbox crossing the antimeridian"""
        config = Config.from_yaml(TEST_CONFIG)
        with caplog.at_level(logging.INFO):
            config.tagging_region = BoundingBox(10, 50, -10, 55)
        assert "coordinates will be rolled" in caplog.text
        assert config.tagging_region.west == 10



"""Tests for entire workflows."""

from pathlib import Path

import matplotlib.colors
import matplotlib.testing.compare
import numpy
import pytest
import xarray as xr
from click.testing import CliRunner
from matplotlib import pyplot as plt

from wam2layers.cli import cli
from wam2layers.config import Config

try:  # Use cartopy if it's available
    import cartopy.feature as cfeature
    from cartopy import crs
except ImportError:
    crs = None
    cfeature = None


CFG_FILE_BACKWARD = Path("tests/test_data/config_rhine.yaml")
CFG_FILE_FORWARD = Path("tests/test_data/config_west_africa.yaml")


## First we define fixtures to help us write the tests more concisely
# Temporary directories that are consistent between tests:
@pytest.fixture(scope="session")
def temp_directory(tmp_path_factory) -> Path:
    return tmp_path_factory.mktemp("data")


@pytest.fixture(scope="session")
def output_dir_backtrack(temp_directory: Path) -> Path:
    output_dir = temp_directory / "output_backtrack"
    output_dir.mkdir()
    return output_dir


@pytest.fixture(scope="session")
def output_dir_forwardtrack(temp_directory: Path) -> Path:
    output_dir = temp_directory / "output_forwardtrack"
    output_dir.mkdir()
    return output_dir


@pytest.fixture(scope="session")
def preprocessed_dir(temp_directory: Path) -> Path:
    return temp_directory / "preprocessed"


@pytest.fixture(scope="session")
def figures_directory(temp_directory: Path) -> Path:
    fig_dir = temp_directory / "figures"
    fig_dir.mkdir()
    return fig_dir


@pytest.fixture(scope="session")
def test_config_backward(
    temp_directory: Path, preprocessed_dir: Path, output_dir_backtrack: Path
) -> str:
    """Return the path to the config for the preprocess, backward, and visualize tests.

    The processed and output data folders of this config are modified to use temporary
    directories.
    """
    cfg = Config.from_yaml(CFG_FILE_BACKWARD)
    cfg.preprocessed_data_folder = preprocessed_dir
    cfg.output_folder = output_dir_backtrack
    tmp_config = temp_directory / "config_rhine.yaml"
    cfg.to_file(tmp_config)
    return str(tmp_config)


@pytest.fixture(scope="session")
def test_config_forward(
    temp_directory: Path, preprocessed_dir: Path, output_dir_forwardtrack: Path
) -> str:
    """Return the path to the config for the forward tracking test(s).

    The output data folder of this config is modified to use a temporary directory.
    """
    cfg = Config.from_yaml(CFG_FILE_FORWARD)
    cfg.output_folder = output_dir_forwardtrack
    tmp_config = temp_directory / "config_west_africa.yaml"
    cfg.to_file(tmp_config)
    return str(tmp_config)


# CLI runs. Session scope means that these fixtures will only run a single time
@pytest.fixture(scope="session")
def preprocess(test_config_backward: str) -> None:
    runner = CliRunner()
    runner.invoke(
        cli, ["preprocess", "era5", test_config_backward], catch_exceptions=False
    )


@pytest.fixture(scope="session")
def backtrack(test_config_backward: str, preprocess) -> None:
    runner = CliRunner()
    runner.invoke(cli, ["track", test_config_backward], catch_exceptions=False)


@pytest.fixture(scope="session")
def visualize_backward(test_config_backward: str, backtrack) -> None:
    runner = CliRunner()
    runner.invoke(
        cli, ["visualize", "output", test_config_backward], catch_exceptions=False
    )


@pytest.fixture(scope="session")
def visualize_forward(test_config_forward: str, forwardtrack) -> None:
    runner = CliRunner()
    runner.invoke(
        cli, ["visualize", "output", test_config_forward], catch_exceptions=False
    )


@pytest.fixture(scope="session")
def forwardtrack(test_config_forward: str) -> None:
    runner = CliRunner()
    runner.invoke(cli, ["track", test_config_forward], catch_exceptions=False)


def plot_difference(
    expected_output: xr.DataArray, output: xr.DataArray, fpath: Path
) -> None:
    """Plot the difference between the expected output and actual output."""
    if not fpath.parent.exists():
        fpath.parent.mkdir()  # Make figures directory if it does not exist yet

    if crs is not None:
        subplot_kw = dict(projection=crs.PlateCarree())
    else:
        subplot_kw = None

    fig, ax = plt.subplots(
        figsize=(16, 10),
        subplot_kw=subplot_kw,
    )
    (output - expected_output).plot.pcolormesh(
        ax=ax,
        x="longitude",
        cmap="RdBu",
        norm=matplotlib.colors.TwoSlopeNorm(vcenter=0),  # normalize cmap so 0 is center
    )
    if cfeature is not None:
        ax.add_feature(cfeature.COASTLINE, linewidth=0.8)
        ax.add_feature(cfeature.BORDERS, linestyle="-", linewidth=0.2)

    ax.set_title(
        f"Absolute difference [mm] between expected and actual '{output.name}' values."
    )
    fig.savefig(fpath, dpi=200)


def different_output_message(fig_path, output_path, expected_path):
    return (
        "Output results are different! Please verify your results. \n"
        f"A difference plot is available under {fig_path}. \n"
        "If you want to keep the new results and include it in your commit, \n"
        "you can update the reference data with the following command: \n"
        f"`cp {output_path} {expected_path}`"
    )


# Tests start now
class TestPreprocess:
    @pytest.fixture(autouse=True)
    def run_preprocess(self, preprocess):
        pass

    def test_preprocess(self, preprocessed_dir: Path, figures_directory: Path):
        output_path = preprocessed_dir / "2022-08-31_fluxes_storages.nc"
        assert output_path.exists()
        # verify outputs
        expected_path = Path(
            "tests/test_data/verify_output/preprocessed_data/2022-08-31_fluxes_storages.nc"
        )
        expected_output = xr.open_dataset(expected_path)
        output = xr.open_dataset(output_path)

        diff_figure = figures_directory / "preprocess_precip_diff.png"
        plot_difference(
            expected_output["precip"].sum("time"),
            output["precip"].sum("time"),
            diff_figure,
        )

        numpy.testing.assert_allclose(
            expected_output["precip"].values,
            output["precip"].values,
            err_msg=different_output_message(diff_figure, output_path, expected_path),
        )

    def test_log_file(self, preprocessed_dir):
        log_files = [i for i in preprocessed_dir.glob("wam2layers_*.log")]
        assert len(log_files) >= 1

    def test_config(self, preprocessed_dir):
        config_path = preprocessed_dir / "config_rhine.yaml"
        assert config_path.exists()


class TestBacktrack:
    @pytest.fixture(autouse=True)
    def run_backtrack(self, backtrack):
        pass

    @pytest.fixture()
    def output_path(self, output_dir_backtrack: Path):
        return output_dir_backtrack / "backtrack_2022-08-31T18-00.nc"

    def test_backtrack_result(self, output_path: Path, figures_directory: Path):
        assert output_path.exists()

        # verify outputs
        expected_path = Path(
            "tests/test_data/verify_output/backtrack_2022-08-31T18-00.nc"
        )
        expected_output = xr.open_dataset(expected_path)
        output = xr.open_dataset(output_path)

        diff_figure = figures_directory / "backtrack_diff.png"
        plot_difference(
            expected_output["e_track"].isel(time=0),
            output["e_track"].isel(time=0),
            diff_figure,
        )

        numpy.testing.assert_allclose(
            expected_output["e_track"].values,
            output["e_track"].values,
            err_msg=different_output_message(diff_figure, output_path, expected_path),
        )

    def test_time_coord(self, output_path: Path):
        expected_time = numpy.datetime64("2022-08-31T18:00:00")
        assert xr.open_dataset(output_path)["time"].to_numpy() == expected_time

    @pytest.mark.parametrize("attr", ("history", "title"))
    def test_dataset_attrs(self, output_path: Path, attr: str):
        output = xr.open_dataset(output_path)
        assert attr in output.attrs

    @pytest.mark.parametrize("attr", ("long_name", "units", "description"))
    def test_variable_attrs(self, output_path: Path, attr: str):
        output = xr.open_dataset(output_path)
        for var in output.data_vars:
            if var not in ["latitude_bnds", "longitude_bnds", "time_bnds"]:
                assert attr in output[var].attrs

    def test_log_file(self, output_dir_backtrack):
        log_files = [i for i in output_dir_backtrack.glob("wam2layers_*.log")]
        len(log_files) >= 1

    def test_config(self, output_dir_backtrack):
        config_path = output_dir_backtrack / "config_rhine.yaml"
        assert config_path.exists()


class TestForwardtrack:
    @pytest.fixture(autouse=True)
    def run_forwardtrack(self, forwardtrack):
        pass

    @pytest.fixture()
    def output_path(self, output_dir_forwardtrack: Path):
        return output_dir_forwardtrack / "forwardtrack_1979-07-01T23-00.nc"

    def test_forwardtrack(self, output_path: Path, figures_directory: Path):
        assert output_path.exists()
        # verify outputs
        expected_path = Path(
            "tests/test_data/verify_output/forwardtrack_1979-07-01T23-00.nc"
        )
        expected_output = xr.open_dataset(expected_path)
        output = xr.open_dataset(output_path)

        for var in ["p_track_upper", "p_track_lower"]:
            diff_figure = figures_directory / f"forwardtrack_diff_{var}.png"
            plot_difference(
                expected_output[var].isel(time=0), output[var].isel(time=0), diff_figure
            )

            numpy.testing.assert_allclose(
                expected_output[var].values,
                output[var].values,
                err_msg=different_output_message(
                    diff_figure, output_path, expected_path
                ),
            )

    def test_time_coord(self, output_path: Path):
        expected_time = numpy.datetime64("1979-07-01T23:00:00.00")
        assert xr.open_dataset(output_path)["time"].to_numpy() == expected_time

    @pytest.mark.parametrize("attr", ("history", "title"))
    def test_dataset_attrs(self, output_path: Path, attr: str):
        output = xr.open_dataset(output_path)
        assert attr in output.attrs

    @pytest.mark.parametrize("attr", ("long_name", "units", "description"))
    def test_variable_attrs(self, output_path: Path, attr: str):
        output = xr.open_dataset(output_path)
        for var in output.data_vars:
            if var not in ["latitude_bnds", "longitude_bnds", "time_bnds"]:
                assert attr in output[var].attrs

    def test_log_file(self, output_dir_forwardtrack):
        log_files = [i for i in output_dir_forwardtrack.glob("wam2layers_*.log")]
        assert len(log_files) >= 1

    def test_config(self, output_dir_forwardtrack):
        config_path = output_dir_forwardtrack / "config_west_africa.yaml"
        assert config_path.exists()


class TestVisualize:
    @pytest.fixture(autouse=True)
    def run_visualize(self, visualize_backward, visualize_forward):
        pass

    def test_visualize_forward(self, output_dir_forwardtrack: Path):
        output_path = output_dir_forwardtrack / "figures" / "cumulative_sources.png"
        assert output_path.exists()
        expected_output = Path(
            "tests/test_data/verify_output/cumulative_sources_fw.png"
        )
        stdout = matplotlib.testing.compare.compare_images(
            expected_output, output_path, tol=5
        )
        if stdout:
            raise ValueError(
                "Output figures are different! Please check the difference in the produced \n"
                " `cumulative_sources-failed-diff.png` figure and verify your results. \n"
                "If you want to keep the new results and include it in your commit, \n"
                "you can update the reference figure with the following command: \n"
                f"`cp {output_path} {expected_output}`"
            )

    def test_visualize_backward(self, output_dir_backtrack: Path):
        output_path = output_dir_backtrack / "figures" / "cumulative_sources.png"
        assert output_path.exists()
        expected_output = Path(
            "tests/test_data/verify_output/cumulative_sources_bw.png"
        )
        stdout = matplotlib.testing.compare.compare_images(
            expected_output, output_path, tol=5
        )
        if stdout:
            raise ValueError(
                "Output figures are different! Please check the difference in the produced \n"
                " `cumulative_sources-failed-diff.png` figure and verify your results. \n"
                "If you want to keep the new results and include it in your commit, \n"
                "you can update the reference figure with the following command: \n"
                f"`cp {output_path} {expected_output}`"
            )

    def test_log_file(self, output_dir_backtrack):
        log_files = [i for i in output_dir_backtrack.glob("wam2layers_*.log")]
        assert len(log_files) >= 2

    def test_config(self, output_dir_backtrack):
        config_path = output_dir_backtrack / "config_rhine.yaml"
        assert config_path.exists()



"""
Utility for appending to netcdf file along unlimited dimension.
Adapted from https://github.com/pydata/xarray/issues/1672#issuecomment-685222909
"""

import netCDF4
import xarray as xr


def _expand_variable(nc_variable, data, expanding_dim, nc_shape, added_size):
    # For time deltas, we must ensure that we use the same encoding as
    # what was previously stored.
    # We likely need to do this as well for variables that had custom
    # econdings too
    if hasattr(nc_variable, "calendar"):
        data.encoding = {
            "units": nc_variable.units,
            "calendar": nc_variable.calendar,
        }
    data_encoded = xr.conventions.encode_cf_variable(data)
    left_slices = data.dims.index(expanding_dim)
    right_slices = data.ndim - left_slices - 1
    nc_slice = (
        (slice(None),) * left_slices
        + (slice(nc_shape, nc_shape + added_size),)
        + (slice(None),) * (right_slices)
    )
    nc_variable[nc_slice] = data_encoded.data


def append_to_netcdf(filename, ds_to_append, expanding_dim="time"):
    with netCDF4.Dataset(filename, mode="a") as nc:
        nc_coord = nc[expanding_dim]
        nc_shape = len(nc_coord)
        added_size = len(ds_to_append[expanding_dim])
        variables, attrs = xr.conventions.encode_dataset_coordinates(ds_to_append)

        for name, data in variables.items():
            if expanding_dim not in data.dims:
                # Nothing to do, data assumed to the identical
                continue

            nc_variable = nc[name]
            _expand_variable(nc_variable, data, expanding_dim, nc_shape, added_size)



import pandas as pd
import xarray as xr


def initialize_time(config, direction="forward"):
    dt = pd.Timedelta(seconds=config.timestep)

    if direction == "forward":
        t0 = pd.Timestamp(config.tracking_start_date)
        th = t0 + dt / 2  # th is "half" time, i.e. between t0 and t1
        t1 = t0 + dt
    elif direction == "backward":
        t1 = pd.Timestamp(config.tracking_end_date)
        th = t1 - dt / 2  # th is "half" time, i.e. between t0 and t1
        t0 = t1 - dt
    else:
        raise ValueError("Direction should be forward or backward")

    return t0, th, t1, dt


def initialize_tagging_region(bbox, lat, lon):
    """Build a new xr.DataArray with ones inside and zeros outside bbox."""
    lat_in_bbox = (bbox.south <= lat) & (lat <= bbox.north)
    unrolled_lon_in_bbox = (bbox.west <= lon) & (lon <= bbox.east)
    rolled_lon_in_bbox = (bbox.west <= lon) | (lon <= bbox.east)
    rolled_domain = bbox.west > bbox.east
    lon_in_bbox = rolled_lon_in_bbox if rolled_domain else unrolled_lon_in_bbox
    tagging_region = xr.where((lat_in_bbox & lon_in_bbox), 1, 0).values
    return tagging_region



import logging
from datetime import datetime as pydt
from functools import lru_cache
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr

from wam2layers import __version__
from wam2layers.config import Config
from wam2layers.preprocessing.shared import (
    accumulation_to_flux,
    add_bounds,
    calculate_humidity,
    insert_level,
    interpolate,
    sortby_ndarray,
)
from wam2layers.preprocessing.xarray_append import append_to_netcdf
from wam2layers.reference.variables import ERA5_INPUT_DATA_ATTRIBUTES

logger = logging.getLogger(__name__)


PREPROCESS_ATTRS = {
    "title": "ERA5 data preprocessed for use in WAM2layers",
    "history": (
        f"created on {pydt.utcnow():%Y-%m-%dT%H:%M:%SZ} "
        f"using wam2layers version {__version__}."
    ),
    "source": (
        "ECMWF Reanalysis v5 (ERA5), "
        "www.ecmwf.int/en/forecasts/dataset/ecmwf-reanalysis-v5"
    ),
    "references": "doi.org/10.5281/zenodo.7010594, doi.org/10.5194/esd-5-471-2014",
    "Conventions": "CF-1.6",
}


@lru_cache(10)
def log_once(logger, msg: str):
    """Keep track of 10 different messages and then warn again.

    Adapted from: https://stackoverflow.com/a/66062313"""
    logger.info(msg)


def load_data(variable, datetime, config):
    """Load data for given variable and date."""
    template = config.filename_template

    # If it's a 4d variable, we need to set the level type
    if variable in ["u", "v", "q"]:
        if config.level_type == "model_levels":
            prefix = "_ml"
        elif config.level_type == "pressure_levels":
            prefix = "_pl"
    # If it's a 3d variable we do not need to set the level type
    else:
        prefix = ""

    # Load data
    filepath = template.format(
        year=datetime.year,
        month=datetime.month,
        day=datetime.day,
        hour=datetime.hour,
        minute=datetime.minute,
        levtype=prefix,
        variable=variable,
    )
    da = xr.open_dataarray(filepath).sel(time=datetime.strftime("%Y%m%d%H%M"))

    if "lev" in da.coords:
        da = da.rename(lev="level")

    # If it's 4d data we want to select a subset of the levels
    if variable in ["u", "v", "q"] and isinstance(config.levels, list):
        return da.sel(level=config.levels)

    return da


def preprocess_precip_and_evap(datetime, config):
    """Load and pre-process precipitation and evaporation."""
    # All incoming units are accumulations (in m) since previous time step
    evap = load_data("e", datetime, config)
    precip = load_data("tp", datetime, config)  # total precipitation

    # Transfer negative (originally positive) values of evap to precip
    precip = np.maximum(precip, 0) + np.maximum(evap, 0)
    precip = precip

    # Change sign convention to all positive,
    evap = np.abs(np.minimum(evap, 0))

    precip = accumulation_to_flux(precip, input_frequency=config.input_frequency)
    evap = accumulation_to_flux(evap, input_frequency=config.input_frequency)
    return precip, evap


def load_era5_ab():
    """Load a and b coefficients."""
    table = Path(__file__).parent / "tableERA5model_to_pressure.csv"
    df = pd.read_csv(table)
    a = df["a [Pa]"].values
    b = df.b.values
    return a, b


def midpoints(x):
    """Linearly interpolate between the values of an array."""
    return (x[1:] + x[:-1]) / 2


def get_edges(era5_modellevels):
    """Get the values of a and b at the edges of a subset of ERA5 modellevels."""
    a, b = load_era5_ab()

    if era5_modellevels == "all":
        return a, b

    # With python indexing starting at 0 and 137 full levels,
    # count from 0-136 instead of 1-137.
    era5_modellevels = [i - 1 for i in era5_modellevels]

    # Calculate the values of a and be at midpoints and extract levels
    a_full = midpoints(a)[era5_modellevels]
    b_full = midpoints(b)[era5_modellevels]

    # Calculate the values of a and b at the edges *between* these levels
    a_edge = midpoints(a_full)
    b_edge = midpoints(b_full)

    # Reinsert the original outer edges
    a_edge = np.block([a[0], a_edge, a[-1]])
    b_edge = np.block([b[0], b_edge, b[-1]])

    return a_edge, b_edge


def get_dp_modellevels(sp, levels):
    """Calculate pressure jump over subset of era5 modellevels."""
    a, b = get_edges(levels)

    # Use dummy level -1 to avoid collision with existing levels
    # first coord will be discarded anyway on taking the diff below.
    a = xr.DataArray(a, coords={"level": np.insert(levels, 0, -1)})
    b = xr.DataArray(b, coords={"level": np.insert(levels, 0, -1)})

    p_edge = a + b * sp

    # Calculate the difference between the pressure levels
    dp = p_edge.diff("level")  # in Pa

    return dp


def get_dp_pressurelevels(q, u, v, ps, qs, us, vs):
    """Get dp with consistent u, v, q for ERA5 pressure level data."""
    p = u.level.broadcast_like(u) * 100  # Pa

    # Insert top of atmosphere values
    # Assume wind at top same as values at lowest pressure, humidity at top 0
    u = insert_level(u, u.isel(level=0), 0)
    v = insert_level(v, v.isel(level=0), 0)
    q = insert_level(q, 0, 0)
    p = insert_level(p, 0, 0)

    # Insert surface level values (at a high dummy pressure value)
    u = insert_level(u, us, 110000)
    v = insert_level(v, vs, 110000)
    q = insert_level(q, qs, 110000)
    p = insert_level(p, ps, 110000)

    # Sort arrays by pressure (ascending)
    u.values = sortby_ndarray(u.values, p.values, axis=1)
    v.values = sortby_ndarray(v.values, p.values, axis=1)
    q.values = sortby_ndarray(q.values, p.values, axis=1)
    p.values = sortby_ndarray(p.values, p.values, axis=1)

    # Insert boundary level values (at a ridiculous dummy pressure value)
    p_boundary = 0.72878581 * np.array(ps) + 7438.803223
    u = insert_level(u, interpolate(p_boundary, p, u), 150000)
    v = insert_level(v, interpolate(p_boundary, p, v), 150000)
    q = insert_level(q, interpolate(p_boundary, p, q), 150000)
    p = insert_level(p, p_boundary, 150000)

    # Sort arrays by pressure once more (ascending)
    u.values = sortby_ndarray(u.values, p.values, axis=1)
    v.values = sortby_ndarray(v.values, p.values, axis=1)
    q.values = sortby_ndarray(q.values, p.values, axis=1)
    p.values = sortby_ndarray(p.values, p.values, axis=1)

    # Reset level coordinate as its values have become meaningless
    nlev = u.level.size
    u = u.assign_coords(level=np.arange(nlev))
    v = v.assign_coords(level=np.arange(nlev))
    q = q.assign_coords(level=np.arange(nlev))
    p = p.assign_coords(level=np.arange(nlev))

    # Calculate pressure jump
    dp = p.diff("level")
    assert np.all(dp >= 0), "Pressure levels should increase monotonically"

    # Interpolate to midpoints
    midpoints = 0.5 * (u.level.values[1:] + u.level.values[:-1])
    dp = dp.assign_coords(level=midpoints)
    u = u.interp(level=midpoints)
    v = v.interp(level=midpoints)
    q = q.interp(level=midpoints)
    p = p.interp(level=midpoints)

    # mask values below surface
    above_surface = p < np.array(ps)[:, None, :, :]
    u = u.where(above_surface)
    v = v.where(above_surface)
    q = q.where(above_surface)
    p = p.where(above_surface)

    return dp, p, q, u, v, p_boundary


def get_input_dates(config):
    """Dates for pre-processing."""
    return pd.date_range(
        start=config.preprocess_start_date,
        end=config.preprocess_end_date,
        freq=config.input_frequency,
    )


def prep_experiment(config_file):
    """Pre-process all data for a given config file.

    This function expects the following configuration settings:

    - preprocess_start_date: formatted as YYYYMMDD, e.g. '20210701'
    - preprocess_end_date: formatted as YYYYMMDD, e.g. '20210716'
    - level_type: either "pressure_levels" or "model_levels"
    - levels: "all" or a list of integers with the desired (model or pressure)
      levels.
    - input_folder: path where raw era5 input data can be found, e.g.
      /home/peter/WAM2layers/era5_2021
    - preprocessed_data_folder: path where preprocessed data should be stored.
      This directory will be created if it does not exist. E.g.
      /home/peter/WAM2layers/preprocessed_data_2021
    - filename_prefix: Fixed part of filename. This function will infer the
      variable name and add _ml for model level data. E.g. with prefix =
      "FloodCase_202107" this function will be able to find
      FloodCase_202107_ml_u.nc or FloodCase_202107_u.nc and
      FloodCase_202107_sp.nc
    """
    config = Config.from_yaml(config_file)

    for datetime in get_input_dates(config):
        is_new_day = datetime == datetime.floor("1d")
        logger.info(datetime)

        # 4d fields
        levels = config.levels
        q = load_data("q", datetime, config)  # in kg kg-1
        u = load_data("u", datetime, config)  # in m/s
        v = load_data("v", datetime, config)  # in m/s
        sp = load_data("sp", datetime, config)  # in Pa

        if config.level_type == "model_levels":
            dp = get_dp_modellevels(sp, levels)

        if config.level_type == "pressure_levels":
            d2m = load_data("d2m", datetime, config)  # Dew point in K
            q2m = calculate_humidity(d2m, sp)  # kg kg-1
            u10 = load_data("u10", datetime, config)  # in m/s
            v10 = load_data("v10", datetime, config)  # in m/s
            dp, p, q, u, v, pb = get_dp_pressurelevels(q, u, v, sp, q2m, u10, v10)

        # Calculate column water vapour
        g = 9.80665  # gravitational accelleration [m/s2]
        cwv = q * dp / g  # (kg/m2)

        try:
            # Calculate column water instead of column water vapour
            tcw = load_data("tcw", datetime, config)  # kg/m2
            correction = tcw / cwv.sum(dim="level")
            cw = correction * cwv  # column water (kg/m2)
            if is_new_day:
                logger.info(
                    f"Total column water correction: mean over grid for this timestep {correction.mean().item():.4f}"
                )

        except FileNotFoundError:
            # Fluxes will be calculated based on the column water vapour
            cw = cwv
            log_once(
                logger, "Total column water not available; using water vapour only"
            )

        # Integrate fluxes and states to upper and lower layer
        if config.level_type == "model_levels":
            # TODO: Check if this is a reasonable choice for boundary
            boundary = 111
            lower_layer = dp.level > boundary
            upper_layer = ~lower_layer

        if config.level_type == "pressure_levels":
            upper_layer = p < pb[:, None, :, :]
            lower_layer = pb[:, None, :, :] < p

        # Vertically integrate state over two layers
        s_lower = cw.where(lower_layer).sum(dim="level")
        s_upper = cw.where(upper_layer).sum(dim="level")

        # Determine the fluxes
        fx = u * cw  # eastward atmospheric moisture flux (kg m-1 s-1)
        fy = v * cw  # northward atmospheric moisture flux (kg m-1 s-1)

        # Vertically integrate fluxes over two layers
        fx_lower = fx.where(lower_layer).sum(dim="level")  # kg m-1 s-1
        fy_lower = fy.where(lower_layer).sum(dim="level")  # kg m-1 s-1
        fx_upper = fx.where(upper_layer).sum(dim="level")  # kg m-1 s-1
        fy_upper = fy.where(upper_layer).sum(dim="level")  # kg m-1 s-1

        precip, evap = preprocess_precip_and_evap(datetime, config)

        # Combine everything into one dataset
        ds = (
            xr.Dataset(
                {
                    "fx_upper": fx_upper.assign_attrs(units="kg m-1 s-1"),
                    "fy_upper": fy_upper.assign_attrs(units="kg m-1 s-1"),
                    "fx_lower": fx_lower.assign_attrs(units="kg m-1 s-1"),
                    "fy_lower": fy_lower.assign_attrs(units="kg m-1 s-1"),
                    "s_upper": s_upper.assign_attrs(units="kg m-2"),
                    "s_lower": s_lower.assign_attrs(units="kg m-2"),
                    "evap": evap,
                    "precip": precip,
                }
            )
            .expand_dims("time")
            .astype("float32")
        )

        # Verify that the data meets all the requirements for the model
        # check_input(ds)

        add_bounds(ds)

        # Add attributes
        for var in ERA5_INPUT_DATA_ATTRIBUTES:
            ds[var].attrs.update(ERA5_INPUT_DATA_ATTRIBUTES[var])
        ds.attrs.update(PREPROCESS_ATTRS)

        # Save preprocessed data
        filename = f"{datetime.strftime('%Y-%m-%d')}_fluxes_storages.nc"
        output_path = config.preprocessed_data_folder / filename

        if is_new_day:
            comp = dict(zlib=True, complevel=8)
            encoding = {var: comp for var in ds.data_vars}
            time_encoding = {"units": "seconds since 1900-01-01"}
            encoding["time"] = time_encoding
            ds.to_netcdf(
                output_path, unlimited_dims=["time"], mode="w", encoding=encoding
            )

        else:
            append_to_netcdf(output_path, ds, expanding_dim="time")



"""
WAM2layers Tagging Region Designer functions

@author: Vincent de Feiter | GitHub: vincentdefeiter

Adapted by Chris Weijenborg

"""
import logging

# Import
import cartopy.crs as ccrs  # used for plotting on a map
import geopandas as gpd  # reading shapefiles
import matplotlib.patheffects as pe  # additional plotting functionalities
import matplotlib.pyplot as plt  # used for plotting purposes
import numpy as np  # used for array calculations
import regionmask  # used for region mapping
import xarray as xr  # used to read .nc files
from cartopy import feature as cfeature

logger = logging.getLogger(__name__)


def mask_with_regionmask(region_source, regions, input_file, output_dir, return_plot):
    """tagging region by using a named region from regionmask.

    This function builds a tagging region based on available regions from the Region Mask Python Package.

    Args:
        region_source (str): indicate here a region source from the Region Mask Python Package (Giorgi, SREX, AR6, PRUDENCE).
        regions (list): single or multiple regions as list.
        reference_file: a reference file of the preprocessing which is used to
            extract the dimensions of the data used.
        output_dir: path where to store the created tagging region.
        return_plot: if true, return a plot that shows the tagging region on a map.

    Returns:
        axes on which the region is plotted if `return_plot` is True

    """

    # Retrieve file dimensions
    file = xr.open_dataset(input_file)

    longitude = np.array(file.longitude)
    latitude = np.array(file.latitude)

    # Set available regions
    if "ar6" in region_source:
        name, region_source = region_source.split(".")
        try:
            available_regions = getattr(regionmask.defined_regions.ar6, region_source)

        except:
            raise KeyError(
                'region_source not specified correctly, please check the "regionmask.defined_regions" options in the Region Mask documentation for more details'
            )
    else:
        try:
            available_regions = getattr(regionmask.defined_regions, region_source)
        except:
            raise KeyError(
                'region_source not specified correctly, please check the "regionmask.defined_regions" options in the Region Mask documentation for more details'
            )

    # create (a) mask(s) of the required region(s):
    mask = getattr(available_regions, "mask_3D")(longitude, latitude)

    # Specify region ID - what to store
    region_ids = np.array(regions)

    if (
        "srex" in region_source
        or "giorgi" in region_source
        or "prudence" in region_source
    ):
        region_ids = region_ids - 1
    if "ocean" in region_source or "land" in region_source:
        raise KeyError(
            "Entry for ar6 not recognized. Please only use the ar6.all options to make the selection. This option ensures proper indexing of the region."
        )

    region_ids = region_ids.tolist()

    if type(region_ids) == int:
        raise KeyError(
            "regions not specified correctly, please check whether your input is provided as a list [...,...], even if it is a single tagging region!"
        )

    new_masks = mask.isel(region=region_ids).values

    # Make WAM2layers fitting
    new_masks = np.where(
        new_masks == False, 0, 1
    )  # Replace True and False by 1 and 0's
    new_masks = np.nanmean(new_masks, axis=0)  # Multiple masks into 1D array

    # Export as tagging region for WAM2layers

    # create xarray dataset
    data = new_masks

    export = xr.Dataset(
        {"tagging_region": (["latitude", "longitude"], data.astype(float))}
    )

    # set coordinates
    export["longitude"] = ("longitude", longitude)
    export["latitude"] = ("latitude", latitude)

    # save to NetCDF file
    export.to_netcdf(output_dir)

    logger.info(f"Stored tagging region in {output_dir}.")

    # Plot as a check
    if return_plot:
        # Visualise all regions available

        fontsizes = 10
        pads = 20

        text_kws = dict(
            bbox=dict(color="none"),
            path_effects=[pe.withStroke(linewidth=2, foreground="w")],
            color="#67000d",
            fontsize=8,
        )

        ax = available_regions.plot(
            projection=ccrs.PlateCarree(), add_ocean=True, text_kws=text_kws
        )

        plt.rcParams["figure.dpi"] = 200  # Set the DPI to 300 (adjust as needed)

        mask_plot = new_masks

        indices = export.where(export.tagging_region > 0, drop=True)

        ax.contourf(
            longitude, latitude, mask_plot, levels=[0.1, 1], zorder=5, alpha=0.8
        )

        if "prudence" in region_source:
            ax.set_extent([-20, 40, 30, 70])

        # Grid
        gl = ax.gridlines(draw_labels=True, alpha=0.5, linestyle=":", color="k")
        gl.top_labels = False
        gl.right_labels = False
        gl.xlabel_style = {"size": fontsizes}
        gl.ylabel_style = {"size": fontsizes}

        return ax


def mask_with_shapefile(shapefile, regions, reference_file, output_dir, return_plot):
    """Mask tagging region using a shapefile.

    This function builds a tagging region based on an existing shapefile and stores the output file.

    Args:
        shapefile: name of the shapefile. Include: .dbf, .prj, .sbn, .sbx, .shp, .shp.xml and .shx files in the input directory.
        regions: if build up of multiple regions, indicate as list (int), otherwise FALSE.
        reference_file: a reference file of the preprocessing which is used to
            extract the dimensions of the data used.
        output_dir: path where to store the created tagging region.
        return_plot: if true, return a plot that shows the tagging region on a map.

    Returns:
        axes on which the region is plotted if 'return_plot' is True.
    """

    # Retrieve file dimensions
    file = xr.open_dataset(reference_file)

    longitude = np.array(file.longitude)
    latitude = np.array(file.latitude)

    # Read shapefile
    ds = gpd.read_file(shapefile)  # load shape file

    # Generate mask
    mask = regionmask.mask_3D_geopandas(ds, longitude, latitude)

    # Specify region ID - what to store
    if regions != False:
        region_ids = np.array(regions)
        region_ids = region_ids.tolist()

        if type(region_ids) == int:
            raise KeyError(
                "regions not specified correctly, please check whether your input is provided as a list [...,...], even if it is a single tagging region!"
            )

        new_masks = mask.isel(region=region_ids).values

        # Make WAM2layers fitting
        new_masks = np.where(
            new_masks == False, 0, 1
        )  # Replace True and False by 1 and 0's
        new_masks = np.nanmean(new_masks, axis=0)  # Multiple masks into 1D array
    else:
        new_masks = mask.isel(region=[0]).values

        # Make WAM2layers fitting
        new_masks = np.where(
            new_masks == False, 0, 1
        )  # Replace True and False by 1 and 0's
        new_masks = np.nanmean(new_masks, axis=0)  # Multiple masks into 1D array

    # Export as tagging region for WAM2layers

    # create xarray dataset
    data = new_masks

    export = xr.Dataset(
        {"tagging_region": (["latitude", "longitude"], data.astype(float))}
    )

    # set coordinates
    export["longitude"] = ("longitude", longitude)
    export["latitude"] = ("latitude", latitude)

    # save to NetCDF file

    export.to_netcdf(output_dir)

    logger.info(f"Stored tagging region in {output_dir}.")

    if return_plot:
        # Visualise all regions available

        fontsizes = 10
        pads = 20

        # Zoomed-in
        fig = plt.figure(figsize=(16, 10), dpi=200)
        ax = fig.add_subplot(111, projection=ccrs.PlateCarree())

        # Make figure
        ax.add_feature(cfeature.LAND, zorder=1, edgecolor="k", facecolor="w")

        ax.add_feature(cfeature.COASTLINE, zorder=2, linewidth=0.8)
        ax.add_feature(cfeature.BORDERS, zorder=2, linewidth=0.2, alpha=0.5)
        ax.add_feature(cfeature.RIVERS, zorder=2, linewidth=3, alpha=0.5)
        ax.add_feature(cfeature.STATES, zorder=2, facecolor="w")
        ax.add_feature(
            cfeature.LAKES,
            zorder=2,
            linewidth=0.8,
            edgecolor="k",
            alpha=0.5,
            facecolor="w",
        )
        mask_plot = new_masks

        indices = export.where(export.tagging_region > 0, drop=True)
        lon_indices = indices.longitude
        lon_indices_c = []
        for i in lon_indices:
            if i <= 180:
                item = i * -1
                lon_indices_c.append(item)
            else:
                lon_indices_c.append(i)

        ax.contourf(
            longitude, latitude, mask_plot, levels=[0.1, 1], zorder=5, alpha=0.8
        )
        ax.set_extent(
            [
                min(lon_indices_c) - 1,
                max(lon_indices_c) + 1,
                indices.latitude.min() - 1,
                indices.latitude.max() + 1,
            ]
        )
        ax.set_title("Zoomed-in", fontsize=fontsizes)

        # Grid
        gl = ax.gridlines(draw_labels=True, alpha=0.5, linestyle=":", color="k")
        gl.top_labels = False
        gl.right_labels = False
        gl.xlabel_style = {"size": fontsizes * 2}
        gl.ylabel_style = {"size": fontsizes * 2}

        # Zoomed-out
        fig = plt.figure(figsize=(16, 10), dpi=200)
        ax2 = fig.add_subplot(111, projection=ccrs.PlateCarree())

        # Make figure
        ax2.add_feature(cfeature.LAND, zorder=1, edgecolor="k", facecolor="w")

        ax2.add_feature(cfeature.COASTLINE, zorder=2, linewidth=0.8)
        ax2.add_feature(cfeature.BORDERS, zorder=2, linewidth=0.2, alpha=0.5)
        ax2.add_feature(cfeature.RIVERS, zorder=2, linewidth=3, alpha=0.5)
        ax2.add_feature(cfeature.STATES, zorder=2, facecolor="w")
        ax2.add_feature(
            cfeature.LAKES,
            zorder=2,
            linewidth=0.8,
            edgecolor="k",
            alpha=0.5,
            facecolor="w",
        )
        mask_plot = new_masks

        indices = export.where(export.tagging_region > 0, drop=True)
        lon_indices = indices.longitude
        lon_indices_c = []
        for i in lon_indices:
            if i <= 180:
                item = i * -1
                lon_indices_c.append(item)
            else:
                lon_indices_c.append(i)

        ax2.contourf(
            longitude, latitude, mask_plot, levels=[0.1, 1], zorder=5, alpha=0.8
        )
        ax2.set_extent(
            [
                min(lon_indices_c) - 20,
                max(lon_indices_c) + 20,
                indices.latitude.min() - 20,
                indices.latitude.max() + 20,
            ]
        )
        ax2.set_title("Zoomed-out", fontsize=fontsizes)

        # Grid
        gl = ax2.gridlines(draw_labels=True, alpha=0.5, linestyle=":", color="k")
        gl.top_labels = False
        gl.right_labels = False
        gl.xlabel_style = {"size": fontsizes * 2}
        gl.ylabel_style = {"size": fontsizes * 2}

        return ax, ax2


def mask_around_point(
    centerpoint,
    radius,
    reference_file,
    output_dir,
    return_plot,
):
    """Mask tagging region using a center point with a given radius.

    This function builds a square tagging region based on given coordinates and
    radius and stores it the output file.

    Args:
        centerpoint: Coordinates of the central point (latitude, longitude).
        radius (int): distance from center to edges of square box in degrees.
        reference_file: a reference file of the preprocessing which is used to
            extract the dimensions of the data used.
        output_dir: path where to store the created tagging region.
        return_plot: if true, return a plot that shows the tagging region on a map.

    Returns:
        axes on which the region is plotted if `return_plot' is True
    """

    # Retrieve file dimensions
    file = xr.open_dataset(reference_file)

    longitude_r = np.array(file.longitude)
    latitude_r = np.array(file.latitude)

    # Create a grid of latitudes and longitudes
    longitude, latitude = np.meshgrid(longitude_r, latitude_r)

    # retrieve latitude and longitude from ceterpoint
    center_lat, center_lon = centerpoint

    # Set the square region of value 1 in the mask
    mask = (np.abs(latitude - center_lat) <= radius) & (
        np.abs(longitude - center_lon) <= radius
    )

    new_masks = np.where(mask == False, 0, 1)  # Replace True and False by 1 and 0's

    # Export as tagging region for WAM2layers

    # create xarray dataset
    data = new_masks

    export = xr.Dataset(
        {"tagging_region": (["latitude", "longitude"], data.astype(float))}
    )

    # set coordinates
    export["longitude"] = ("longitude", longitude_r)
    export["latitude"] = ("latitude", latitude_r)

    # save to NetCDF file
    export.to_netcdf(output_dir)

    logger.info(f"Stored tagging region in {output_dir}.")

    if return_plot:
        # Visualise all regions available

        fontsizes = 10
        pads = 20

        fig = plt.figure(figsize=(16, 10), dpi=200)
        ax = fig.add_subplot(111, projection=ccrs.PlateCarree())

        # Make figure
        ax.add_feature(cfeature.LAND, zorder=1, edgecolor="k", facecolor="w")

        ax.add_feature(cfeature.COASTLINE, zorder=2, linewidth=0.8)
        ax.add_feature(cfeature.BORDERS, zorder=2, linewidth=0.2, alpha=0.5)
        ax.add_feature(cfeature.RIVERS, zorder=2, linewidth=3, alpha=0.5)
        ax.add_feature(cfeature.STATES, zorder=2, facecolor="w")
        ax.add_feature(
            cfeature.LAKES,
            zorder=2,
            linewidth=0.8,
            edgecolor="k",
            alpha=0.5,
            facecolor="w",
        )
        mask_plot = new_masks

        ax.contourf(longitude_r, latitude_r, mask_plot, levels=[0.1, 1])

        indices = export.where(export.tagging_region > 0, drop=True)
        ax.contourf(
            longitude, latitude, mask_plot, levels=[0.1, 1], zorder=5, alpha=0.8
        )
        ax.set_extent(
            [
                indices.longitude.min() - 10,
                indices.longitude.max() + 10,
                indices.latitude.min() - 10,
                indices.latitude.max() + 10,
            ]
        )

        ax.plot(
            center_lon,
            center_lat,
            marker="o",
            markerfacecolor="r",
            markeredgecolor="k",
            linestyle="None",
            markersize=radius * 0.5,
            zorder=6,
        )

        # Grid
        gl = ax.gridlines(draw_labels=True, alpha=0.5, linestyle=":", color="k")
        gl.top_labels = False
        gl.right_labels = False
        gl.xlabel_style = {"size": fontsizes * 2}
        gl.ylabel_style = {"size": fontsizes * 2}
        return ax


def mask_around_track(
    centerpoints,
    times,
    radius,
    reference_file,
    output_file,
    return_plot,
):
    """Mask tagging region using a track of center points with a given radius (currently in degrees).

    This function builds a square source region based on given coordinates and
    radius and stores it the output file.

    Args:
        centerpoints: List of coordinates of the central points, changing over time [(lat1, lon1), ....,lat_final, lon_final)] .
        radius (int): distance from center to edges of square box in degrees.
        reference_file: a reference file of the preprocessing which is used to
            extract the dimensions of the data used.
        output_file: path where to store the created source region.
        return_plot: if true, return a plot that shows the source region on a map.

    Returns:
        axes on which the region is plotted if `return_plot' is True
    """

    # Retrieve file dimensions
    file = xr.open_dataset(reference_file)

    longitude_r = np.array(file.longitude)
    latitude_r = np.array(file.latitude)

    # Create a grid of latitudes and longitudes
    longitude, latitude = np.meshgrid(longitude_r, latitude_r)

    # Initiate result array
    new_masks = np.zeros((len(centerpoints), len(latitude_r), len(longitude_r)))

    # loop over times
    for t in range(len(centerpoints)):
        centerpoint = centerpoints[t]
        # retrieve latitude and longitude from ceterpoint
        center_lat, center_lon = centerpoint

        # Set the square region of value 1 in the mask
        mask = (np.abs(latitude - center_lat) <= radius) & (
            np.abs(longitude - center_lon) <= radius
        )

        new_masks[t, ::] = np.where(
            mask == False, 0, 1
        )  # Replace True and False by 1 and 0's

    # Export as source region for WAM2layers
    # create xarray dataset
    data = new_masks

    export = xr.Dataset(
        {"source_region": (["time", "latitude", "longitude"], data.astype(float))}
    )

    # set coordinates
    export["longitude"] = ("longitude", longitude_r)
    export["latitude"] = ("latitude", latitude_r)
    export["time"] = ("time", times)

    # save to NetCDF file
    export.to_netcdf(output_file)

    logger.info(f"Stored source region in {output_file}.")

    if return_plot:
        # Visualise all regions available

        fontsizes = 10
        pads = 20

        fig = plt.figure(figsize=(16, 10), dpi=200)
        ax = fig.add_subplot(111, projection=ccrs.PlateCarree())

        # Make figure
        ax.add_feature(cfeature.LAND, zorder=1, edgecolor="k", facecolor="w")

        ax.add_feature(cfeature.COASTLINE, zorder=2, linewidth=0.8)
        ax.add_feature(cfeature.BORDERS, zorder=2, linewidth=0.2, alpha=0.5)
        ax.add_feature(cfeature.RIVERS, zorder=2, linewidth=3, alpha=0.5)
        ax.add_feature(cfeature.STATES, zorder=2, facecolor="w")
        ax.add_feature(
            cfeature.LAKES,
            zorder=2,
            linewidth=0.8,
            edgecolor="k",
            alpha=0.5,
            facecolor="w",
        )
        mask_plot = new_masks

        ax.contourf(longitude_r, latitude_r, mask_plot, levels=[0.1, 1])

        indices = export.where(export.source_region == 1, drop=True)
        ax.contourf(
            longitude, latitude, mask_plot, levels=[0.1, 1], zorder=5, alpha=0.8
        )
        ax.set_extent(
            [
                indices.longitude.min() - 10,
                indices.longitude.max() + 10,
                indices.latitude.min() - 10,
                indices.latitude.max() + 10,
            ]
        )

        ax.plot(
            center_lon,
            center_lat,
            marker="o",
            markerfacecolor="r",
            markeredgecolor="k",
            linestyle="None",
            markersize=radius * 0.5,
            zorder=6,
        )

        # Grid
        gl = ax.gridlines(draw_labels=True, alpha=0.5, linestyle=":", color="k")
        gl.top_labels = False
        gl.right_labels = False
        gl.xlabel_style = {"size": fontsizes * 2}
        gl.ylabel_style = {"size": fontsizes * 2}
        return ax



import logging
import time

import numpy as np
import pandas as pd
import psutil
import xarray as xr

from wam2layers.config import Config

logger = logging.getLogger(__name__)


class Profiler:
    def __init__(self):
        self.t0 = time.time()
        self.mem0 = self._current_memory()

    def __call__(self):
        """Report memory and time increments."""
        dt = round((time.time() - self.t0), 2)
        dmem = self._current_memory() - self.mem0
        # self.t0 = time.time()
        # self.mem0 = self._current_memory()
        return dt, dmem

    def _current_memory(self):
        return psutil.Process().memory_info().rss / (1024 * 1024)


class ProgressTracker:
    def __init__(self, output, mode="backtrack"):
        """Keep track of tagged and tracked moisture."""
        self.mode = mode
        self.total_tagged_moisture = 0
        self.tracked = 0
        self.lost_water = 0
        self.gained_water = 0
        self.boundary_transport = 0
        self.store_intermediate_states(output)
        self.profile = Profiler()
        self.stability_correction_previous_grid = 0.0
        self.stability_correction_previous_factor = 1.0

    def store_intermediate_states(self, output):
        """Update moisture totals based on output since previous time step.

        Intended for when the output fields are reset at write time, so we
        don't lose accumulations from previous outputs.
        """
        totals = output.sum()

        if self.mode == "backtrack":
            self.tracked += totals["e_track"]
            self.total_tagged_moisture += totals["tagged_precip"]
        else:  # mode is forwardtrack
            self.tracked += totals["p_track_lower"] + totals["p_track_upper"]
            self.total_tagged_moisture += totals["tagged_evap"]

        # Don't include boundary losses in log messages
        interior_losses = output.losses[1:-1, 1:-1].sum()
        boundary_transport = totals["losses"] - interior_losses

        self.lost_water += interior_losses
        self.boundary_transport += boundary_transport
        self.gained_water += totals["gains"]

    def print_progress(self, t, output):
        """Print some useful runtime diagnostics."""
        totals = output.sum()

        if self.mode == "backtrack":
            tracked = self.tracked + totals["e_track"]
            total_tagged_moisture = self.total_tagged_moisture + totals["tagged_precip"]
        else:  # mode is forwardtrack
            tracked = self.tracked + totals["p_track_upper"] + totals["p_track_lower"]
            total_tagged_moisture = self.total_tagged_moisture + totals["tagged_evap"]
        still_in_atmosphere = (
            totals["s_track_upper_restart"] + totals["s_track_lower_restart"]
        )

        # Don't include boundary losses in log messages
        interior_losses = output.losses[1:-1, 1:-1].sum()
        boundary_transport = totals["losses"] - interior_losses

        lost_water = self.lost_water + interior_losses
        boundary_transport = self.boundary_transport + boundary_transport
        gained_water = self.gained_water + totals["gains"]

        tracked_percentage = tracked / total_tagged_moisture * 100
        in_atmos_percentage = still_in_atmosphere / total_tagged_moisture * 100
        lost_percentage = lost_water / total_tagged_moisture * 100
        gained_percentage = gained_water / total_tagged_moisture * 100
        boundary_percentage = boundary_transport / total_tagged_moisture * 100
        closure_check_percentage = (
            tracked_percentage
            + in_atmos_percentage
            + lost_percentage
            + gained_percentage
            + boundary_percentage
        )

        time, memory = self.profile()

        # TODO: print boundary losses and internal losses separately
        logger.info(
            f"{t} - "
            f"Tracked moisture: {tracked_percentage.item():.2f}%. "
            f"Tagged in atmosphere: {in_atmos_percentage.item():.2f}%. "
            f"Boundary transport: {boundary_percentage.item():.2f}%. "
            f"Lost moisture: {lost_percentage.item():.2f}%. "
            f"Gained moisture: {gained_percentage.item():.2f}%. "
            f"Tagging closure: {closure_check_percentage.item():.2f}%. "
            f"Time since start: {time}s, RAM: {memory:.2f} MB"
        )

    def track_stability_correction(
        self,
        fy_corrected: xr.DataArray,
        fy_abs: xr.DataArray,
        config: Config,
        t: pd.Timestamp,
    ) -> None:
        """Issue warning if correction exceeds criterion.

        Warning advises to reduce the timestep.

        Criteria:
        1. A correction is applied for more than 5% of the grid (first time)
        2. At least 10% more of the grid is corrected compared to the last warning
        3. Previous strongest correction factor is exceeded by 10%
        """
        corrected_area = fy_corrected < fy_abs
        corrected_percent = corrected_area.sum() / corrected_area.count() * 100
        correction_factor = fy_corrected / fy_abs
        strongest_correction = correction_factor.min()

        warn = False
        if strongest_correction < self.stability_correction_previous_factor / 1.1:
            self.stability_correction_previous_factor = strongest_correction
            warn = True

        elif (
            corrected_percent > 5  # at least 5 percent of the grid should be corrected
            and corrected_percent > self.stability_correction_previous_grid * 1.1
        ):
            self.stability_correction_previous_grid = corrected_percent
            warn = True

        if warn:
            # Write correction field to output debug file
            # TODO: doesn't feel like the right place to do this.
            # a model object could improve the code structure.
            debug_dir = config.output_folder / "debug"
            debug_dir.mkdir(exist_ok=True)
            timestamp = t.strftime("%Y%m%d-%H%M%S")
            filename = debug_dir / f"stability_correction_{timestamp}.nc"

            correction_factor.name = "correction_factor"
            correction_factor.to_netcdf(filename)

            logger.warning(
                f"Warning: stability correction applied to {corrected_percent:.1f}% of "
                "grid.\n"
                f"    Average correction factor was {correction_factor.mean():.2f}, "
                f"the strongest correction factor was {strongest_correction:.2f}\n"
                f"    The correction factor field is written to {filename}.\n"
                "    The next warning will only be raised when at least 10% more cells "
                "are corrected, or the strongest correction factor is 10% stronger."
            )



import numpy as np
import pytest
import xarray as xr

from wam2layers.config import BoundingBox
from wam2layers.tracking.io import select_subdomain


@pytest.fixture
def ds():
    lon = np.arange(-180, 180, 30)
    lat = np.arange(90, -90, -30)
    time = np.arange(3)

    p = xr.DataArray(
        np.random.randn(len(time), len(lat), len(lon)),
        coords={"time": time, "latitude": lat, "longitude": lon},
    )
    e = xr.DataArray(
        np.random.randn(len(time), len(lat), len(lon)),
        coords={"time": time, "latitude": lat, "longitude": lon},
    )
    return xr.Dataset({"e": e, "p": p})


def test_select_subdomain(ds):
    """Check subdomain that fits in [-180, 180]."""
    bbox = BoundingBox(0, -20, 90, 60)
    sub = select_subdomain(ds, bbox)
    assert np.array_equal(sub.latitude, [60, 30, 0])
    assert np.array_equal(sub.longitude, [0, 30, 60, 90])


def test_rolling_subdomain(ds):
    """Check subdomain that does not fit in [-180, 180] or [0, 360]."""
    bbox = BoundingBox(130, -20, 40, 60)
    sub = select_subdomain(ds, bbox)
    assert np.array_equal(sub.longitude, [150, -180, -150, -120, -90, -60, -30, 0, 30])



import warnings
from pathlib import Path

import numpy as np
import yaml


def _warning_on_one_line(message, category, filename, lineno, line=None):
    """A more compact warning format.

    https://stackoverflow.com/a/26433913
    """
    short_path = Path(filename).relative_to(Path(filename).parents[0])
    return f"{short_path}:{lineno}: {category.__name__}: {message}\n"


# Override the format with which warnings are printed
warnings.formatwarning = _warning_on_one_line


def check_monotonic_increase(data, dim="level"):
    """Check if data is monotonically increasing in the given dimension."""
    if not data.diff(dim).min() >= 0:
        warnings.warn(
            f"Data is not monotonically increasing in the dimension {dim} for {data.name}."
        )


def check_monotonic_decrease(data, dim="level"):
    """Check if data is monotonically decreasing in the given dimension."""
    if not data.diff(dim).max() <= 0:
        warnings.warn(
            f"Data is not monotonically decreasing in the dimension {dim} for {data.name}."
        )


def check_units(data, valid_units):
    """Check if data has the given units."""
    if "units" not in data.attrs:
        warnings.warn(f"{data.name} has no units")
    elif data.units not in valid_units:
        warnings.warn(
            f"{data.name} has units '{data.units}'; expected (one of) {valid_units}."
        )


def check_positive(data):
    """Check if data is positive."""
    if not data.min() >= 0:
        warnings.warn(f"Data for {data.name} is not positive")


def check_negative(data):
    """Check if data is negative."""
    if not data.max() <= 0:
        warnings.warn(f"Data for {data.name} is not negative.")


def check_nonnegative(data):
    """Check if data is nonnegative."""
    if not data.min() >= 0:
        warnings.warn(f"Data for {data.name} is not nonnegative.")


def check_order_of_magnitude(data, order):
    """Check if data is within the given order of magnitude."""
    if not data.min() >= 10 ** order and data.max() <= 10**order:
        warnings.warn(
            f"Data  for {data.name}is not within the order of magnitude {order}."
        )


def check_frequency(data, frequency):
    """Check if data has the given frequency."""
    if not data.time.dt.frequency == frequency:
        warnings.warn(
            f"Data  for {data.name} has frequency {data.time.dt.frequency} instead of {frequency}."
        )


def check_shape(data, ndim):
    if data.ndim != ndim:
        warnings.warn("Expected 3 dimensions, found {data.ndim} for {data.name}.")


def check_range(data, range):
    if not data.min() >= range[0] and data.max() <= range[1]:
        warnings.warn(f"Data for {data.name} is not within the range {min} to {max}.")


def check_uniform(coord):
    spacing = np.diff(coord)
    if not spacing.min() == spacing.max():
        warnings.warn(
            f"Coordinate spacing is not uniform for coord {coord} of {data.name}"
        )


def check_coords(data, coords):
    check_shape(data, ndim=len(coords))
    for coord in coords:
        if coord not in data.coords:
            warnings.warn(f"Data for {data.name} does not have a coordinate {coord}.")
        else:
            check_uniform(data[coord])
        if coord == "latitude":
            check_monotonic_decrease(data[coord], dim=coord)
        else:
            check_monotonic_increase(data[coord], dim=coord)


def check_input(data):
    for variable, var_info in VARIABLES.items():
        check_units(data[variable], var_info["valid_units"])
        check_coords(data[variable], var_info["coordinates"])
        check_range(data[variable], var_info["range"])
        if var_info["all_positive"]:
            check_positive(data[variable])


VARIABLES = yaml.safe_load(
    """
    evap:
        valid_units: ["kg m-2 s-1", "kg/m2/s"]
        range: [0, 0.001]
        all_positive: True
        coordinates: [time, latitude, longitude]
    precip:
        valid_units: ["kg m-2 s-1", "kg/m2/s"]
        range: [0, 0.1]
        all_positive: True
        coordinates: [time, latitude, longitude]
    s_upper:
        valid_units: ["kg m-2", "kg/m2"]
        range: [0, 1000]
        all_positive: True
        coordinates: [time, latitude, longitude]
    s_lower:
        valid_units: ["kg m-2", "kg/m2"]
        range: [0, 1000]
        all_positive: True
        coordinates: [time, latitude, longitude]
    fx_upper:
        valid_units: ["kg m-1 s-1", "kg m^-1 s^-1", "kg m**-1 s**-1"]
        range: [-1000, 1000]
        all_positive: False
        coordinates: [time, latitude, longitude]
    fy_upper:
        valid_units: ["kg m-1 s-1", "kg m^-1 s^-1", "kg m**-1 s**-1"]
        range: [-1000, 1000]
        all_positive: False
        coordinates: [time, latitude, longitude]
    fx_lower:
        valid_units: ["kg m-1 s-1", "kg m^-1 s^-1", "kg m**-1 s**-1"]
        range: [-1000, 1000]
        all_positive: False
        coordinates: [time, latitude, longitude]
    fy_lower:
        valid_units: ["kg m-1 s-1", "kg m^-1 s^-1", "kg m**-1 s**-1"]
        range: [-1000, 1000]
        all_positive: False
        coordinates: [time, latitude, longitude]
"""
)



"""Holds reference information on variables (names, units, etc.)."""

DATA_ATTRIBUTES: dict[str, dict[str, str]] = {
    "e_track": {
        "long_name": "tracked evaporation",
        "units": "kg m-2 accumulated per output time step ",
        "description": "evaporation that will end up as tagged precipitation",
    },
    "s_track_upper_restart": {
        "long_name": "tracked atmospheric moisture storage state in upper layer",
        "units": "kg m-2",
        "description": (
            "tracked atmospheric moisture storage state in upper layer exactly at the "
            "out time step (initial condition, i.e., 'restart' for next timestep)"
        ),
    },
    "s_track_lower_restart": {
        "long_name": "tracked atmospheric moisture storage state in lower layer",
        "units": "kg m-2",
        "description": (
            "tracked atmospheric moisture storage state in lower layer exactly at the "
            "out time step (initial condition, i.e., 'restart' for next timestep)"
        ),
    },
    "p_track_upper": {
        "long_name": "tracked precipitation in upper layer",
        "units": "kg m-2 accumulated per output time step",
        "description": (
            "precipitation in the upper layer that originates from tagged evaporation"
        ),
    },
    "p_track_lower": {
        "long_name": "tracked precipitation in lower layer ",
        "units": "kg m-2 accumulated per output time step",
        "description": (
            "precipitation in the lower layer that originates from tagged evaporation"
        ),
    },
    "tagged_evap": {
        "long_name": "tagged evaporation ",
        "units": "kg m-2 accumulated per output time step",
        "description": (
            "tagged evaporation in the tagging region, i.e., input for the tracking"
        ),
    },
    "tagged_precip": {
        "long_name": "tagged precipitation ",
        "units": "kg m-2 accumulated per output time step",
        "description": (
            "tagged precipitation in the tagging region, i.e., input for the tracking"
        ),
    },
    "losses": {
        "long_name": "moisture lost during the tracking",
        "units": "kg m-2 accumulated per output time step ",
        "description": (
            "moisture that is 'lost' either internally (due to s_track > s) or simply "
            "fluxes over the boundary of then domain"
        ),
    },
    "gains": {
        "long_name": "moisture gained during the tracking",
        "units": "kg m-2 accumulated per output time step",
        "description": (
            "moisture that is numerically gained due to violating stability criterion "
            "(if this is variable is significant something goes wrong and debugging "
            "is strongly recommended!)"
        ),
    },
    "latitude": {
        "long_name": "latitude",
        "units": "degrees_north",
    },
    "longitude": {
        "long_name": "longitude",
        "units": "degrees_east",
    },
    "latitude_bnds": {
        "long_name": "latitude bounds",
        "units": "degrees_north",
        "description": "Left and right bounds of the grid cells.",
    },
    "longitude_bnds": {
        "long_name": "longitude bounds",
        "units": "degrees_east",
        "description": "Top and bottom bounds of the grid cells",
    },
}


ERA5_INPUT_DATA_ATTRIBUTES: dict[str, dict[str, str]] = {
    "fx_upper": {
        "long_name": "eastward moisture flux in the upper layer",
        "units": "kg m-1 s-1",
        "description": (
            "Vertical integral of eastward moisture flux in the upper atmospheric "
            "layer. Instantaneous parameter (i.e. valid at time coordinate)."
        ),
    },
    "fx_lower": {
        "long_name": "eastward moisture flux in the lower layer",
        "units": "kg m-1 s-1",
        "description": (
            "Vertical integral of eastward moisture flux in the lower atmospheric "
            "layer. Instantaneous parameter (i.e. valid at time coordinate)."
        ),
    },
    "fy_upper": {
        "long_name": "northward moisture flux in the upper layer",
        "units": "kg m-1 s-1",
        "description": (
            "Vertical integral of northward moisture flux in the upper atmospheric "
            "layer. Instantaneous parameter (i.e. valid at time coordinate)."
        ),
    },
    "fy_lower": {
        "long_name": "northward moisture flux in the lower layer",
        "units": "kg m-1 s-1",
        "description": (
            "Vertical integral of northward moisture flux in the lower atmospheric "
            "layer. Instantaneous parameter (i.e. valid at time coordinate)."
        ),
    },
    "s_upper": {
        "long_name": "total moisture in the upper layer",
        "units": "kg m-2",
        "description": (
            "Vertical integral of moisture in the upper atmospheric layer. "
            "Instantaneous parameter (i.e. valid at time coordinate)."
        ),
    },
    "s_lower": {
        "long_name": "total moisture in the lower layer",
        "units": "kg m-2",
        "description": (
            "Vertical integral of moisture in the lower atmospheric layer. "
            "Instantaneous parameter (i.e. valid at time coordinate)."
        ),
    },
    "evap": {
        "long_name": "Moisture flux from the surface into the atmosphere",
        "units": "kg m-2 s-1",
        "description": (
            "The total moisture flux from the surface to the surface. "
            "Accumulated parameter: accumulated from the previous time coordinate to "
            "the current time coordinate. "
            "The values of this variable are only positive."
        ),
    },
    "precip": {
        "long_name": "Moisture flux from the atmosphere to the surface",
        "units": "kg m-2 s-1",
        "description": (
            "The total moisture flux from the atmosphere to the surface. "
            "Note that this includes all precipitation, as well as dew formation. "
            "Accumulated parameter: accumulated from the previous time coordinate to "
            "the current time coordinate. "
            "The values of this variable are only positive."
        ),
    },
    "latitude": {
        "long_name": "latitude",
        "units": "degrees_north",
    },
    "longitude": {
        "long_name": "longitude",
        "units": "degrees_east",
    },
    "latitude_bnds": {
        "long_name": "latitude bounds",
        "units": "degrees_north",
        "description": "Left and right bounds of the grid cells.",
    },
    "longitude_bnds": {
        "long_name": "longitude bounds",
        "units": "degrees_east",
        "description": "Top and bottom bounds of the grid cells",
    },
}



"""Command-line interface for the WAM2layers model.

This module contains all the functionality that makes it possible to run wam2layers on the command line, like so:

    wam2layers preprocess era5 floodcase.yaml
    wam2layers track floodcase.yaml

et cetera. It is built with [click](https://click.palletsprojects.com/en/8.1.x/).
"""

import logging
import shutil
import sys
from datetime import datetime
from pathlib import Path
from typing import Union

import click

from wam2layers import __version__
from wam2layers.analysis import visualization
from wam2layers.config import Config
from wam2layers.preprocessing.era5 import prep_experiment
from wam2layers.tracking.backtrack import run_experiment as run_backtrack_experiment
from wam2layers.tracking.forwardtrack import (
    run_experiment as run_forwardtrack_experiment,
)

logger = logging.getLogger(__name__)


def setup_logging(log_path: Union[str, Path], debug: bool):
    """Configure logging behaviour.

    Messages with logging.INFO (and higher) are written to stdout
    Messages with logging.DEBUG (and higher) are written to wam2layers.log

    Args:
        log_path: Folder in which the logging file should be created.
        debug: If DEBUG level messages should also be printed to the terminal.
    """
    # https://stackoverflow.com/a/58828499
    timestamp = datetime.now().strftime("%Y%m%d-%H%M%S")

    file_handler = logging.FileHandler(
        Path(log_path, f"wam2layers_{timestamp}.log"), mode="w"
    )
    stream_handler = logging.StreamHandler()

    if debug:
        stream_handler.setLevel(logging.DEBUG)
    else:
        stream_handler.setLevel(logging.INFO)

    logging.basicConfig(
        level=logging.DEBUG,
        format="%(message)s",
        handlers=[file_handler, stream_handler],
    )


def _copy_config_yaml(yaml_path, target_path):
    """Copy config file to the output path."""
    shutil.copy(yaml_path, target_path)


def get_wam_version():
    """Get the version of WAM2layers"""
    pkg_dir = Path(__file__).parent.absolute()
    return (
        f"wam2layers {__version__} from {pkg_dir} "
        f"(python {sys.version_info.major}.{sys.version_info.minor})"
    )


@click.group()
@click.pass_context
@click.version_option(message=get_wam_version())
@click.option("--debug", is_flag=True, help="Print debug statements to terminal.")
def cli(ctx, debug):
    """Command line interface to WAM2layers."""
    ctx.obj = dict()
    ctx.obj["debug"] = debug


# Command line setup for tracking
@cli.command()
@click.pass_context
@click.argument("config_file", type=click.Path(exists=True))
def track(ctx, config_file):
    """Run WAM2layers tracking experiment.

    CONFIG_FILE: Path to WAM2layers experiment configuration file.

    Usage examples:

        \b
        - wam2layers track path/to/cases/era5_2021.yaml
    """
    config = Config.from_yaml(config_file)
    log_path = config.output_folder
    setup_logging(log_path, ctx.obj["debug"])
    _copy_config_yaml(config_file, log_path)
    logger.info("Welcome to WAM2layers.")
    logger.info(f"Starting {config.tracking_direction} tracking experiment.")
    if config.tracking_direction == "backward":
        run_backtrack_experiment(config_file)
    else:
        run_forwardtrack_experiment(config_file)


@cli.command(hidden=True)
def backtrack():
    msg = (
        "The `backtrack` and `forwardtrack` commands have been removed in favor of "
        "`track`.\nPlease specify the tracking direction in your config file and use "
        "the `track` command instead."
        "\n"
        "\nFor example, in your config.yaml file add:"
        "\n    # Tracking"
        "\n    tracking_direction: backward"
    )
    raise ValueError(msg)


@cli.command(hidden=True)
def forwardtrack():
    msg = (
        "The `backtrack` and `forwardtrack` commands have been removed in favor of "
        "`track`.\nPlease specify the tracking direction in your config file and use "
        "the `track` command instead."
        "\n"
        "\nFor example, in your config.yaml file add:"
        "\n    # Tracking"
        "\n    tracking_direction: forward"
    )
    raise ValueError(msg)


# Command line setup for preprocess
@click.group()
def preproc_cli():
    """Pre-process raw input data for tracking with WAM2layers"""
    pass


@preproc_cli.command()
@click.pass_context
@click.argument("config_file", type=click.Path(exists=True))
def era5(ctx, config_file):
    """Preprocess ERA5 data for WAM2layers tracking experiments.

    CONFIG_FILE: Path to WAM2layers experiment configuration file.

    Usage examples:

        \b
        - wam2layers preprocess era5 path/to/cases/era5_2021.yaml
    """
    log_path = Config.from_yaml(config_file).preprocessed_data_folder
    setup_logging(log_path, ctx.obj["debug"])
    _copy_config_yaml(config_file, log_path)
    logger.info("Welcome to WAM2layers.")
    logger.info("Starting preprocessing ERA5 data.")
    prep_experiment(config_file)


# Command line setup for visualization
@click.group()
def visualize_cli():
    """Visualize input or output data of a WAM2layers experiment"""
    ...


@visualize_cli.command()
@click.pass_context
@click.argument("config_file", type=click.Path(exists=True))
def input(ctx, config_file):
    """Visualize input data for experiment.

    CONFIG_FILE: Path to WAM2layers experiment configuration file.

    Usage examples:

        \b
        - wam2layers visualize input path/to/cases/era5_2021.yaml
    """
    log_path = Config.from_yaml(config_file).output_folder
    setup_logging(log_path, ctx.obj["debug"])
    _copy_config_yaml(config_file, log_path)
    logger.info("Welcome to WAM2layers.")
    logger.info("Starting visualizing input data.")
    visualization.visualize_input_data(config_file)


@visualize_cli.command()
@click.pass_context
@click.argument("config_file", type=click.Path(exists=True))
def output(ctx, config_file):
    """Visualize output data for experiment.

    CONFIG_FILE: Path to WAM2layers experiment configuration file.

    Usage examples:

        \b
        - wam2layers visualize output path/to/cases/era5_2021.yaml
    """
    log_path = Config.from_yaml(config_file).output_folder
    setup_logging(log_path, ctx.obj["debug"])
    _copy_config_yaml(config_file, log_path)
    logger.info("Welcome to WAM2layers.")
    logger.info("Starting visualizing output data.")
    visualization.visualize_output_data(config_file)


@visualize_cli.command()
@click.pass_context
@click.argument("config_file", type=click.Path(exists=True))
def both(ctx, config_file):
    """Visualize both input and output data for experiment."""
    log_path = Config.from_yaml(config_file).output_folder
    setup_logging(log_path, ctx.obj["debug"])
    _copy_config_yaml(config_file, log_path)
    logger.info("Welcome to WAM2layers.")
    logger.info("Starting visualizing both input and output data.")
    visualization.visualize_both(config_file)


cli.add_command(preproc_cli, name="preprocess")
cli.add_command(visualize_cli, name="visualize")


if __name__ == "__main__":
    cli()



import logging
from pathlib import Path

import numpy as np
import pandas as pd
import xarray as xr

from wam2layers.config import Config
from wam2layers.tracking.core import (
    calculate_fz,
    horizontal_advection,
    stabilize_fluxes,
    vertical_advection,
    vertical_dispersion,
)
from wam2layers.tracking.io import (
    load_data,
    load_tagging_region,
    output_path,
    write_output,
)
from wam2layers.tracking.shared import initialize_tagging_region, initialize_time
from wam2layers.utils.grid import get_boundary, get_grid_info, stagger_x, stagger_y
from wam2layers.utils.profiling import ProgressTracker

logger = logging.getLogger(__name__)


def backtrack(
    F,
    S1,
    S0,
    tagging_mask,
    output,
    config,
):
    # Unpack input data
    fx_upper = stagger_x(F["fx_upper"].values)
    fy_upper = stagger_y(F["fy_upper"].values)
    fx_lower = stagger_x(F["fx_lower"].values)
    fy_lower = stagger_y(F["fy_lower"].values)
    evap = F["evap"].values
    precip = F["precip"].values
    f_vert = F["f_vert"].values
    s_upper = S1["s_upper"].values
    s_lower = S1["s_lower"].values

    s_track_upper = output["s_track_upper_restart"].values
    s_track_lower = output["s_track_lower_restart"].values

    # Short name for often used expressions
    s_track_relative_lower = np.minimum(s_track_lower / s_lower, 1.0)
    s_track_relative_upper = np.minimum(s_track_upper / s_upper, 1.0)

    # Actual tracking (note: backtracking, all fluxes change sign)
    bc = config.periodic_boundary  # boundary condition True/False
    # TODO: apply terms in successive steps instead of all at once?
    s_track_lower += config.timestep * (
        +horizontal_advection(s_track_relative_lower, -fx_lower, -fy_lower, bc)
        + vertical_advection(-f_vert, s_track_relative_lower, s_track_relative_upper)
        + vertical_dispersion(
            -f_vert, s_track_relative_lower, s_track_relative_upper, config.kvf
        )
        + tagging_mask * precip * s_lower / (s_upper + s_lower)
        - evap * s_track_relative_lower
    )

    s_track_upper += config.timestep * (
        +horizontal_advection(s_track_relative_upper, -fx_upper, -fy_upper, bc)
        - vertical_advection(-f_vert, s_track_relative_lower, s_track_relative_upper)
        - vertical_dispersion(
            -f_vert, s_track_relative_lower, s_track_relative_upper, config.kvf
        )
        + tagging_mask * precip * s_upper / (s_upper + s_lower)
    )

    # lower and upper: redistribute unaccounted water that is otherwise lost from the sytem
    overshoot_lower = np.maximum(0, s_track_lower - S0["s_lower"])
    overshoot_upper = np.maximum(0, s_track_upper - S0["s_upper"])
    s_track_lower = s_track_lower - overshoot_lower + overshoot_upper
    s_track_upper = s_track_upper - overshoot_upper + overshoot_lower

    # account for negative storages that are set to zero: "numerically gained water"
    gains_lower = np.where(s_track_lower < 0, -s_track_lower, 0)
    gains_upper = np.where(s_track_upper < 0, -s_track_upper, 0)
    gains = gains_lower + gains_upper
    s_track_lower = np.maximum(s_track_lower, 0)
    s_track_upper = np.maximum(s_track_upper, 0)

    # At this point any of the storages could still be overfull, thus stabilize and assigns losses
    losses_lower = np.maximum(0, s_track_lower - S0["s_lower"])
    losses_upper = np.maximum(0, s_track_upper - S0["s_upper"])
    losses = losses_lower + losses_upper
    s_track_lower = np.minimum(s_track_lower, S0["s_lower"])
    s_track_upper = np.minimum(s_track_upper, S0["s_upper"])

    # Bookkeep boundary transport as "lost moisture at grid edges"
    boundary = get_boundary(losses, periodic=config.periodic_boundary)
    losses += xr.where(boundary, (s_track_upper + s_track_lower), 0)
    s_track_upper = xr.where(boundary, 0, s_track_upper)
    s_track_lower = xr.where(boundary, 0, s_track_lower)

    # Update output fields
    output["tagged_precip"] += tagging_mask * precip * config.timestep
    output["e_track"] += evap * s_track_relative_lower * config.timestep
    output["s_track_lower_restart"].values = s_track_lower
    output["s_track_upper_restart"].values = s_track_upper
    output["losses"] += losses
    output["gains"] += gains


def initialize(config_file):
    """Read config, region, and initial states."""
    logger.info(f"Initializing experiment with config file {config_file}")

    config = Config.from_yaml(config_file)

    # Initialize outputs as empty fields based on the input coords
    t = pd.Timestamp(config.tracking_end_date)
    grid = load_data(t, config, "states").coords
    output = xr.Dataset(
        {
            # Keep last state for a restart
            "s_track_upper_restart": xr.DataArray(0, coords=grid).astype("float32"),
            "s_track_lower_restart": xr.DataArray(0, coords=grid).astype("float32"),
            "e_track": xr.DataArray(0, coords=grid).astype("float32"),
            "tagged_precip": xr.DataArray(0, coords=grid).astype("float32"),
            "losses": xr.DataArray(0, coords=grid).astype("float32"),
            "gains": xr.DataArray(0, coords=grid).astype("float32"),
        }
    )

    if config.restart:
        # Reload last state from existing output
        date = config.tracking_end_date
        ds = xr.open_dataset(output_path(date, config))
        output["s_track_upper_restart"].values = ds.s_track_upper_restart.values
        output["s_track_lower_restart"].values = ds.s_track_lower_restart.values

    logger.info(f"Output will be written to {config.output_folder.absolute()}.")
    return config, output, grid


#############################################################################
# With the correct import statements, the code in the function below could
# alternatively be be used as a script in a separate python file or notebook.
#############################################################################


def run_experiment(config_file):
    """Run a backtracking experiment from start to finish."""
    config, output, grid = initialize(config_file)

    progress_tracker = ProgressTracker(output, mode="backtrack")

    t0, th, t1, dt = initialize_time(config, direction="backward")

    if isinstance(config.tagging_region, Path):
        tagging_region = load_tagging_region(config)
        tagging_region_stationary = tagging_region.ndim == 2
    else:
        lat, lon = grid["latitude"], grid["longitude"]
        bbox = config.tagging_region
        tagging_region = initialize_tagging_region(bbox, lat, lon)
        tagging_region_stationary = True

    while t0 >= config.tracking_start_date:
        S0 = load_data(t0, config, "states")
        F = load_data(th, config, "fluxes")
        S1 = load_data(t1, config, "states")

        # We need the gradient of the flux, so divide by grid spacing
        a, dy, dx = get_grid_info(F)
        for level in ["upper", "lower"]:
            # Convert to accumulations for budget calculations
            F["fx_" + level] /= dx
            F["fy_" + level] /= dy

        # Load/update tagging mask
        tagging_mask = 0
        if config.tagging_start_date <= t1 <= config.tagging_end_date:
            if tagging_region_stationary:
                tagging_mask = tagging_region
            else:
                tagging_mask = load_tagging_region(config, t=t1)

        # Apply a stability correction if needed
        stabilize_fluxes(F, S1, progress_tracker, config, t1)

        # Determine the vertical moisture flux
        F["f_vert"] = calculate_fz(F, S0, S1, config.timestep, config.kvf)

        # Inside backtrack the "output" dictionary is updated
        backtrack(F, S1, S0, tagging_mask, output, config)
        t1 -= dt
        th -= dt
        t0 -= dt

        # Daily output
        is_output_time = t1 == t1.floor(config.output_frequency)
        is_final_step = t0 < config.tracking_start_date
        if is_output_time or is_final_step:
            progress_tracker.print_progress(t1, output)
            progress_tracker.store_intermediate_states(output)
            write_output(output, t1, config)
            # Flush previous outputs
            output[["e_track", "tagged_precip", "losses", "gains"]] *= 0

    logger.info("Experiment complete.")



"""Functionality related to the grid."""

import numpy as np
import xarray as xr


def stagger_x(f):
    """Interpolate f to the grid cell interfaces.
    Only the values at the interior interfaces are returned
    Arguments:
        f: 2d array of shape [M, N]
    Returns:
        2d array of shape [M-2, N-1]
    """
    return 0.5 * (f[:, :-1] + f[:, 1:])[1:-1, :]


def stagger_y(f):
    """Interpolate f to the grid cell interfaces.
    Only the values at the interior interfaces are returned
    Arguments:
        f: 2d array of shape [M, N]
    Returns:
        2d array of shape [M-1, N-2]
    """
    return 0.5 * (f[:-1, :] + f[1:, :])[:, 1:-1]


def get_grid_info(ds):
    """Return grid cell area and lenght of the sides."""
    dg = 111089.56  # [m] length of 1 degree latitude
    erad = 6.371e6  # [m] Earth radius

    latitude = ds.latitude.values
    longitude = ds.longitude.values
    dx_deg = np.abs(longitude[1] - longitude[0])  # [degrees]
    dy_deg = np.abs(latitude[1] - latitude[0])  # [degrees]

    # Calculate area TODO check this calculation!
    lat_n = np.minimum(90.0, latitude + 0.5 * dx_deg)
    lat_s = np.maximum(-90.0, latitude - 0.5 * dx_deg)

    a = (
        np.pi
        / 180.0
        * erad**2
        * dx_deg
        * abs(np.sin(lat_s * np.pi / 180.0) - np.sin(lat_n * np.pi / 180.0))
    )

    # Calculate faces
    dy = dy_deg * dg  # [m] grid spacing in meridional direction
    dx_n_gridcell = dx_deg * dg * np.cos((latitude + dx_deg / 2) * np.pi / 180)
    dx_s_gridcell = dx_deg * dg * np.cos((latitude - dx_deg / 2) * np.pi / 180)
    dx = 0.5 * (dx_n_gridcell + dx_s_gridcell)  # [m] grid spacing in zonal direction
    return a, dy, dx[:, None]


def get_boundary(field, periodic=False):
    """Return a mask with 1 along the boundary and 0 in the interior."""
    boundary = xr.ones_like(field, dtype=bool)
    # Mask interior
    if periodic:
        boundary[1:-1, 0] = 0
    else:
        boundary[1:-1, 1:-1] = 0

    return boundary



import importlib
import logging
import warnings
from pathlib import Path

import cmocean
import matplotlib.pyplot as plt
import pandas as pd
import xarray as xr

from wam2layers.config import Config
from wam2layers.tracking.io import input_path, load_tagging_region, output_path

logger = logging.getLogger(__name__)


MISSING_CARTOPY_MSG = (
    "Cartopy is not available, plots may look different than expected.\n"
    "To install the optional visualization packages do `pip install wam2layers[viz]`."
)


def package_available(package_name: str) -> bool:
    """Return true if the specified package installed in the python kernel."""
    if importlib.util.find_spec(package_name) is not None:
        return True
    return False


def get_projection():
    """Get the cartopy 'plate carree' projection if available, else return None"""
    if package_available("cartopy"):
        import cartopy.crs

        return cartopy.crs.PlateCarree()
    else:
        warnings.warn(MISSING_CARTOPY_MSG, UserWarning, stacklevel=1)
        return None


def polish(ax, region):
    if package_available("cartopy"):
        import cartopy.feature

        ax.add_feature(cartopy.feature.COASTLINE, linewidth=0.8)
        ax.add_feature(cartopy.feature.BORDERS, linestyle="-", linewidth=0.2)
    else:
        warnings.warn(MISSING_CARTOPY_MSG, UserWarning, stacklevel=1)
    ax.set_xlim(region.longitude.min(), region.longitude.max())
    ax.set_ylim(region.latitude.min(), region.latitude.max())


def _plot_input(config: Config, ax):
    """
    Return subplot with input.
    backtrack = precipitation
    forwardtrack = evaporation
    """
    # Load config and some useful stuf.
    region = load_tagging_region(config)

    # Load data
    start = config.tagging_start_date
    end = config.tagging_end_date
    dates = pd.date_range(
        start=start,
        end=end,
        freq=config.output_frequency,
        inclusive="left",
    )

    input_files = []
    for date in dates[:-1]:
        input_files.append(input_path(date, config))
    ds = xr.open_mfdataset(input_files, combine="nested", concat_dim="time")
    if config.tracking_direction == "backward":
        subset = ds.precip.sel(time=slice(start, end))
    else:
        subset = ds.evap.sel(time=slice(start, end))

    # TODO: check if backtrack files or forwardtrack files were in output
    input = (subset * region * 3600).sum("time").compute()

    # Make figure
    input.plot(ax=ax, cmap=cmocean.cm.rain, cbar_kwargs=dict(fraction=0.05, shrink=0.5))
    ax.set_title("Cumulative input during tagging [mm]", loc="left")
    polish(ax, region.where(region > 0, drop=True))


def _plot_output(config: Config, ax):
    """
    Return subplot with tracked moisture
    forwardtrack = precipitation
    backtrack = evaporation
    """
    region = load_tagging_region(config)

    # Load data
    dates = pd.date_range(
        start=config.tracking_start_date,
        end=config.tracking_end_date,
        freq=config.output_frequency,
        inclusive="left",
    )

    if config.tracking_direction == "backward":
        output_files = [output_path(date, config) for date in dates[:-1]]
    else:
        output_files = [output_path(date, config) for date in dates[1:]]

    if not all([Path(file).exists() for file in output_files]):
        raise FileNotFoundError(f"Could not find all files: {output_files}.")
    elif len(output_files) == 0:
        raise ValueError("Too few output files to visualize.")

    ds = xr.open_mfdataset(output_files, combine="nested", concat_dim="time")
    if config.tracking_direction == "backward":
        out_track = ds.e_track.sum("time")
    else:
        out_track = ds.p_track_lower.sum("time") + ds.p_track_upper.sum("time")

    # Make figure
    out_track.plot(
        ax=ax,
        vmin=0,
        robust=True,
        cmap=cmocean.cm.rain,
        cbar_kwargs=dict(fraction=0.05, shrink=0.5),
    )

    ax.set_title("Accumulated tracked moisture [mm]", loc="left")

    # Add source region outline
    # TODO: fix this, because no region contours are shown
    region.plot.contour(ax=ax, levels=[1], colors="k")
    polish(ax, region)


def visualize_input_data(config_file):
    """An figure showing the cumulative moisture inputs.

    TODO: make figure creation independent of case.
    """
    config = Config.from_yaml(config_file)

    # Make figure
    fig = plt.figure(figsize=(16, 10))
    ax = fig.add_subplot(111, projection=get_projection())

    _plot_input(config, ax)

    # Save
    out_dir = Path(config.output_folder) / "figures"
    out_dir.mkdir(exist_ok=True, parents=True)
    path_to_fig = out_dir / "input_event.png"
    fig.savefig(path_to_fig, dpi=200)
    logger.info(f"Figure of cumulative moisture inputs written to {path_to_fig}.")


def visualize_output_data(config_file):
    """An figure showing the cumulative moisture origins.

    TODO: make figure creation independent of case.
    """
    # Load config and some useful stuf.
    config = Config.from_yaml(config_file)

    # Make figure
    fig = plt.figure(figsize=(16, 10))
    ax = fig.add_subplot(111, projection=get_projection())
    _plot_output(config, ax)

    # Save
    out_dir = Path(config.output_folder) / "figures"
    out_dir.mkdir(exist_ok=True, parents=True)
    path_to_fig = out_dir / "cumulative_sources.png"
    fig.savefig(path_to_fig, dpi=200)
    logger.info(f"Figure of cumulative moisture origins written to {path_to_fig}.")


def visualize_both(config_file):
    """Diagnostic figure with four subplots combining input and output data."""
    # Load config and some useful stuf.
    config = Config.from_yaml(config_file)

    # Make figure
    fig, [ax1, ax2] = plt.subplots(
        2, 1, figsize=(16, 10), subplot_kw=dict(projection=get_projection())
    )

    _plot_input(config, ax1)
    _plot_output(config, ax2)

    # Save
    out_dir = Path(config.output_folder) / "figures"
    out_dir.mkdir(exist_ok=True, parents=True)
    path_to_fig = out_dir / "summary_subplots.png"
    fig.savefig(path_to_fig, dpi=200)
    logger.info(f"Diagnostic figure of input and output data written to {path_to_fig}.")



import logging
from datetime import datetime
from pathlib import Path
from typing import Dict, List, Literal, NamedTuple, Optional, Union

import yaml
from pydantic import (
    AfterValidator,
    BaseModel,
    ConfigDict,
    Field,
    FilePath,
    field_validator,
    model_validator,
)
from typing_extensions import Annotated

logger = logging.getLogger(__name__)


class BoundingBox(NamedTuple):
    west: Annotated[float, Field(ge=-180, le=180)]
    south: Annotated[float, Field(ge=-80, le=80)]
    east: Annotated[float, Field(ge=-180, le=180)]
    north: Annotated[float, Field(ge=-80, le=80)]


def validate_region(region):
    """Check if region is existing path or valid bbox."""
    if isinstance(region, Path):
        return region.absolute()

    assert -180 <= region.west <= 180, "longitude should be between -180 and 180"
    assert -180 <= region.east <= 180, "longitude should be between -180 and 180"
    assert -80 <= region.north <= 80, "latitude should be between -80 and 80"
    assert -80 <= region.south <= 80, "latitude should be between -80 and 80"
    assert region.south < region.north, "south should be smaller than north"
    if region.west > region.east:
        logger.info("west > east, coordinates will be rolled around meridian")
    return region


class Config(BaseModel):
    # Pydantic configuration
    model_config = ConfigDict(validate_assignment=True)

    # Configuration settings
    filename_template: str
    """The filename pattern of the raw input data.

    Used to find the input files during preprocessing. The pattern will be
    interpreted during execution of the model to find the input data for each
    date and variable.

    For example, the following pattern:

    .. code-block:: yaml

        filename_template: /ERA5data/{year}/{month:02}/ERA5_{year}-{month:02d}-{day:02d}{levtype}_{variable}.nc

    will be converted to

    .. code-block:: python

        /ERA5data/2021/07/ERA5_2021-07-15_ml_u.nc

    for date 2022-07-15, variable u, and levtype "_ml" (note the underscore).
    """

    preprocessed_data_folder: Path
    """Location where the pre-processed data should be stored.

    If it does not exist, it will be created during pre-processing.

    For example:

    .. code-block:: yaml

        preprocessed_data_folder: ~/floodcase_202107/preprocessed_data

    """

    tracking_direction: Literal["forward", "backward"]
    """The tracking direction, either forward or backward.

    You have to specify if either forward or backward tracking should be performed.

    For example:

    .. code-block:: yaml

        tracking_direction: backward

    """

    tagging_region: Annotated[
        Union[FilePath, BoundingBox], AfterValidator(validate_region)
    ]
    """Subdomain delimiting the source/sink regions for tagged moisture.

    You can either specify a path that contains a netcdf file, or a bounding box
    of the form [west, south, east, north].

    The bounding box should be inside -180, -80, 180, 80; if west > south, the
    coordinates will be rolled to retain a continous longitude.

    The file should exist. If it has a time dimension, the nearest field will be
    used as tagging region, and the time should still be between
    tagging_start_date and tagging_end_date

    For example:

    .. code-block:: yaml

        tagging_region: /data/volume_2/era5_2021/tagging_region_global.nc
        tagging_region: [0, 50, 10, 55]
    """

    tracking_domain: Optional[
        Annotated[BoundingBox, AfterValidator(validate_region)]
    ] = None
    """Subdomain delimiting the region considered during tracking.

    This is useful when you have global pre-processed data but you don't need
    global tracking.

    You can specify a bounding box of the form [west, south, east, north].

    The bounding box should be inside -180, -80, 180, 80; if west > south, the
    coordinates will be rolled to retain a continous longitude.

    If it is set to `null`, then it will use full domain of preprocessed data.

    Note that you should always set period to False if you use a subdomain.

    For example:

    .. code-block:: yaml

        tracking_domain: [0, 50, 10, 55]

    """

    output_folder: Path
    """Location where output of tracking and analysis should be written.

    For example:

    .. code-block:: yaml

        output_folder: ~/floodcase_202107/output_data

    """

    preprocess_start_date: datetime
    """Start date for preprocessing.

    Should be formatted as: `"YYYY-MM-DD[T]HH:MM"`. Start date < end date.
    The preprocess_start_date is included in the preprocessing.

    For example:

    .. code-block:: yaml

        preprocess_start_date: "2021-07-01T00:00"

    """

    preprocess_end_date: datetime
    """End date for preprocessing.

    Should be formatted as: `"YYYY-MM-DD[T]HH:MM"`. Start date < end date.
    The preprocess_end_date is included in the preprocessing.

    For example:

    .. code-block:: yaml

        preprocess_end_date: "2021-07-15T23:00"
    """

    tracking_start_date: datetime
    """Start date for tracking.

    Should be formatted as: `"YYYY-MM-DD[T]HH:MM"`. Start date < end date, even if
    backtracking.
    When backward tracking the tracking_start_date is not given as output date.

    For example:

    .. code-block:: yaml

        tracking_start_date: "2021-07-01T00:00"

    """

    tracking_end_date: datetime
    """Start date for tracking.

    Should be formatted as: `"YYYY-MM-DD[T]HH:MM"`. Start date < end date, even if
    backtracking.

    For example:

    .. code-block:: yaml

        tracking_end_date: "2021-07-15T23:00"
    """

    tagging_start_date: datetime
    """Start date for tagging.

    For tracking individual (e.g. heavy precipitation) events, you can set the
    start and end date to something different than the total tracking start and
    end date, you can also indicate the hours that you want to track. The
    start date is included.

    Should be formatted as: `"YYYY-MM-DD[T]HH:MM"`. Start date < end date, even if
    backtracking.

    For example:

    .. code-block:: yaml

        tagging_start_date: "2021-07-13T00:00"

    """

    tagging_end_date: datetime
    """End date for tagging.

    For tracking individual (e.g. heavy precipitation) events, you can set the
    start and end date to something different than the total tracking start and
    end date, you can also indicate the hours that you want to track. The
    end date is included.

    Should be formatted as: `"YYYY-MM-DD[T]HH:MM"`. Start date < end date, even if
    backtracking.

    For example:

    .. code-block:: yaml

        tagging_end_date: "2021-07-14T23:00"

    """

    input_frequency: str
    """Frequency of the raw input data.

    Used to calculated water volumes.

    For example:

    .. code-block:: yaml

        input_frequency: '1h'

    """

    timestep: int
    """Timestep in seconds with which to perform the tracking.

    The data will be interpolated during model execution. Too large timestep will
    violate CFL criterion, too small timestep will lead to excessive numerical
    diffusion and slow progress. For best performance, the input frequency
    should be divisible by the timestep.

    For example:

    .. code-block:: yaml

        timestep: 600  # timestep in seconds

    """

    output_frequency: str
    """Frequency at which to write output to file.

    For example, for daily output files:

    .. code-block:: yaml

        output_frequency: '1d'

    """

    level_type: Literal["model_levels", "pressure_levels"]
    """Type of vertical levels in the raw input data.

    Can be either `model_levels` or `pressure_levels`.

    For example:

    .. code-block:: yaml

        level_type: model_levels

    """

    levels: "Union[List[int], Literal['All']]"
    """Which levels to use from the raw input data.

    A list of integers corresponding to the levels in the input data, or a
    subset thereof. Shorthand `"all"` will attempt to use all 137 ERA5 levels.

    For example:

    .. code-block:: yaml

        levels: [20,40,60,80,90,95,100,105,110,115,120,123,125,128,130,131,132,133,134,135,136,137]

    """

    restart: bool
    """Whether to restart from previous run.

    If set to `true`, this will attempt to read the output from a previous model
    run and continue from there. The output from the previous timestep must be
    available for this to work.

    For example:

    .. code-block:: yaml

        restart: False

    """

    periodic_boundary: bool
    """Whether to use period boundaries in the zonal direction.

    This should be used when working with global datasets.

    For example:

    .. code-block:: yaml

        periodic_boundary: true
    """

    kvf: float
    """net to gross vertical flux multiplication parameter

    For example:

    .. code-block:: yaml

        kvf: 3

    """

    @classmethod
    def from_yaml(cls, config_file):
        """Read settings from a configuration.yaml file.

        For example:

        .. highlight:: python
        .. code-block:: python

            from wam2layers.config import Config
            config = Config.from_yaml('../../cases/floodcase_2021.yaml')

        """
        with open(config_file) as f:
            settings = yaml.safe_load(f)

        return cls(**settings)

    @field_validator("preprocessed_data_folder", "output_folder")
    @classmethod
    def _expanduser(cls, path):
        """Resolve ~ in input paths."""
        return path.absolute()

    @field_validator("preprocessed_data_folder", "output_folder")
    @classmethod
    def _make_dir(cls, path):
        """Create output dirs if they don't exist yet."""
        if not path.exists():
            logger.info(f"Creating output folder {path}")
            path.mkdir(parents=True)
        return path

    @model_validator(mode="after")
    def check_date_order(self):
        if self.tracking_start_date > self.tracking_end_date:
            raise ValueError(
                "tracking_end_date should be later than tracking_start_date"
            )
        if self.tagging_start_date > self.tagging_end_date:
            raise ValueError("tagging_end_date should be later than tagging_start_date")
        if self.preprocess_start_date > self.preprocess_end_date:
            raise ValueError(
                "preprocess_end_date should be later than preprocess_start_date"
            )

        if self.tracking_domain is not None and self.periodic_boundary:
            logger.warning(
                "Periodic boundary set to true while using a subdomain. Are you sure?"
            )

        return self

    def to_file(self, fname: Union[str, Path]) -> None:
        """Export the configuration to a file.

        Note that any comments and formatting from an original yaml file is lost.
        """
        fpath = Path(fname)
        with fpath.open("w") as f:
            yaml.dump(self.model_dump(mode="json"), f)
